/*                                              Somfai Ellak, 1992.feb.
Ez a file nehany jo  es (az erdekesseg kedveert) nehany kozismerten rossz
rnd generator C implementaciojat tartalmazza.
A fuggvenyek formaja:
seed..(long seed);    seed adas  0 es RAN.._MAX-1 kozott
long  lran..();       long pszeudorandom szam  0 es RAN.._MAX-1 kozott
double ran..();       double pszeudorandom szam  0 es 1 kozott
 
 
A generatorokon vegzett nehany teszt eredmenye:
 
          speed   2^n_corr   3D_kitoltes   gyakorisag_khi^2    futam
ranlin     12.4      ok          ok             ok               ok
ranks       5.0      ok          ok           rossz              ok
ranksz      6.0      ok          ok             ok               ok
ran3        3.8      ok          ok             ok               ok?
randu       7.6    rossz        rossz           ok              rossz
ransub      5.5      ok          ok             ok               ok?
 
A speed teszt egy 33MHz-es 486-on Turbo C 2.0 compilerrel keszult,
a szamok egy egyszeru ertekadas idejet mutatjak mikrosec-ban (double).
2^n korrelaltsagot a 2^n modusu lin kongruencia generatorok mutatnak.
A gyakorisag-probat 10000x10000-es, a futamprobat 1000x30000 mintakon
vegeztem.
 
A lin kongruencia generartorok hasznalata meg akkor sem ajanlott, ha konkretan
ezeket a teszteket kielegitettek. A ranks a gyakorisagproban esett ki, igy a
RANKSZ es a RAN3 / RANSUB generatorok maradtak. (Bar az utobbi ketto gyenge
eredmenyt mutatott a futamproban, csak ez nem volt eleg szignifikans.)
*/
 
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
 
 
  /* ***********************************************************
   * Linear Congruence Rnd Generator                           *
   * r[n+1] = (16807 * r[n]) mod 2^31-1                        *
   * in: R.M.Ziff: An observed bias in the Linear.. (unpubl.)  *
   *********************************************************** */
 
#define RANLIN_MAX 2147483648.0
long lin_num;
 
seedlin (seed) 
long seed;
{
  if(seed<0) { puts("SEEDLIN error."); exit(1); }
  lin_num=seed;
  }
 
long lranlin () {
  lin_num=(long)fmod(16807.0*(double)lin_num,2147483647.0 );
  return(lin_num);
  }
 
double ranlin () {
  lin_num=(long)fmod(16807.0*(double)lin_num,2147483647.0 );
  return((double)lin_num/RANLIN_MAX);
  }
 
 
  /* *****************************************************************
   * Kirkpatrick-Stoll Linear Recurrence Rnd Generator               *
   * n++                                                             *
   * r[n] = r[n-103] xor r[n-250]                                    *
   * out -> r[n]                                                     *
   * in: ... R.M.Ziff: An Observed Bias in the Linear..   (unpubl.)  *
   *     !!! NOT RECOMENDED !!!                                      *
   ***************************************************************** */
 
#define RANKS_MAX 2147483648.0
long ks_num[256];
int ks_i;
 
seedks  (seed) 
long seed;
{
  seedlin(seed);
  if((seed<0)||(seed>=RANKS_MAX)) { puts("SEEDKS error."); exit(1); }
  for(ks_i=0;ks_i<256;ks_i++) ks_num[ks_i]=lranlin();
  }
 
long  lranks  () {
  ks_i= ++ks_i & 255;
  ks_num[ks_i]=ks_num[(ks_i-103)&255]^ks_num[(ks_i-250)&255];
  return(ks_num[ks_i]);
  }
 
double ranks () {
  ks_i= ++ks_i & 255;
  ks_num[ks_i]=ks_num[(ks_i-103)&255]^ks_num[(ks_i-250)&255];
  return((double)ks_num[ks_i]/RANKS_MAX);
  }
 
 
 
  /* ********************************************************************
   * Modified Kirkpatrick-Stoll Linear Recurrence Rnd Generator by Ziff *
   * n++                                                                *
   * r[n] = r[n-103] xor r[n-250]                                       *
   * n++                                                                *
   * r[n] = r[n- 30] xor r[n-127]                                       *
   * out -> r[n]                                                        *
   * in: R.M.Ziff: An Observed Bias in the Linear..   (unpubl.)         *
   ******************************************************************** */
 
#define RANKSZ_MAX 2147483648.0
long ksz_num[256];
int ksz_i;
 
seedksz  (seed) 
long seed;
{
  if(seed<0) { puts("SEEDKSZ error."); exit(1); }
  seedlin(seed);
  for(ksz_i=0;ksz_i<256;ksz_i++) ksz_num[ksz_i]=lranlin();
  }
 
long  lranksz  () {
  ksz_i= ++ksz_i & 255;
  ksz_num[ksz_i]=ksz_num[(ksz_i-103)&255]^ksz_num[(ksz_i-250)&255];
  ksz_i= ++ksz_i & 255;
  ksz_num[ksz_i]=ksz_num[(ksz_i-30)&255]^ksz_num[(ksz_i-127)&255];
  return(ksz_num[ksz_i]);
  }
 
double ranksz () {
  ksz_i= ++ksz_i & 255;
  ksz_num[ksz_i]=ksz_num[(ksz_i-103)&255]^ksz_num[(ksz_i-250)&255];
  ksz_i= ++ksz_i & 255;
  ksz_num[ksz_i]=ksz_num[(ksz_i-30)&255]^ksz_num[(ksz_i-127)&255];
  return((double)ksz_num[ksz_i]/RANKSZ_MAX);
  }
 
 
 
  /* ******************************************************
   * "RAN3" Substractive Rnd Generator                    *
   * n++                                                  *
   * r[n] = r[n-55] - r[n-24]                             *
   * in: Numerical Recipes  p.199                         *
   ****************************************************** */
 
#define RAN3_MAX 1000000000L
#define RAN3_MSEED 161803398L
#define RAN3_MZ 0
#define RAN3_FAC 1.E-9
 
long ran3_ma[55];
int  ran3_inext,ran3_inextp;
 
seed3  (seed) 
long seed;
{
  long mj,mk;
  int i,ii,k;
 
  mj=RAN3_MSEED-(seed<0)?(-seed):seed;
  mj=mj-mj/RAN3_MAX*RAN3_MAX;
  ran3_ma[54]=mj;
  mk=1;
  for(i=1;i<55;i++) {
    ii=21*i-(21*i)/55*55;
    ran3_ma[ii]=mk;
    mk=mj-mk;
    if(mk<RAN3_MZ) mk=mk+RAN3_MAX;
    mj=ran3_ma[ii];
    }
  for(k=0;k<4;k++)
   for(i=0;i<55;i++) {
     ran3_ma[i]=ran3_ma[i]-ran3_ma[1+ i+30-(i+30)/55*55];
     if(ran3_ma[i]<RAN3_MZ) ran3_ma[i]=ran3_ma[i]+RAN3_MAX;
     }
   ran3_inext=0;
   ran3_inextp=31;
}
 
long   lran3  () {
  long mj;
  ran3_inext++;
  if(ran3_inext==56) ran3_inext=1;
  ran3_inextp++;
  if(ran3_inextp==56) ran3_inextp=1;
  mj=ran3_ma[ran3_inext]-ran3_ma[ran3_inextp];
  if(mj<RAN3_MZ) mj=mj+RAN3_MAX;
  ran3_ma[ran3_inext]=mj;
  return(mj);
  }
 
double ran3 () {
  long mj;
  ran3_inext++;
  if(ran3_inext==56) ran3_inext=1;
  ran3_inextp++;
  if(ran3_inextp==56) ran3_inextp=1;
  mj=ran3_ma[ran3_inext]-ran3_ma[ran3_inextp];
  if(mj<RAN3_MZ) mj=mj+RAN3_MAX;
  ran3_ma[ran3_inext]=mj;
  if(mj>RAN3_MAX) puts("RAN3 error.");
  return(mj*RAN3_FAC);
  }
 
 
 
 
 
  /* ***************************************************
   * RANDU Linear Congruence Rnd Generator             *
   * r[n+1] = (65539 * r[n]) mod 2^31                  *
   * in: Knuth: The Art of Computer programming        *
   * !!! NOT RECOMENDED !!!                            *
   *************************************************** */
 
#define RANDU_MAX 2147483648.0
long du_num;
 
seeddu (seed) 
long seed;
{
  if(seed<0) { puts("SEEDDU error."); exit(1); }
  du_num=seed | 1;
  }
 
long lrandu () {
  du_num=2147483647L & (long) (65539.0*(double)du_num);
  return(du_num);
  }
 
double randu () {
  du_num=2147483647L & (long) (65539.0*(double)du_num);
  return((double)du_num/RANDU_MAX);
  }
 
 
 
 
  /* ********************************
   *  Substractive Rnd Generator    *
   *  Similar to RAN3               *
   ******************************** */
 
#define RANSUB_MAX 2147483648.0
long ransub_ma[55];
int  ransub_inext,ransub_inextp;
 
seedsub  (seed) 
long seed;
{
  int i;
  if(seed<0) { puts("SEEDSUB error."); exit(1); }
  seedlin(seed);
  for(i=1;i<55;i++) ransub_ma[i]=lranlin();
  ransub_inext=0;
  ransub_inextp=31;
}
 
long   lransub  () {
  long mj;
  ransub_inext++;
  if(ransub_inext==56) ransub_inext=1;
  ransub_inextp++;
  if(ransub_inextp==56) ransub_inextp=1;
  mj=ransub_ma[ransub_inext]-ransub_ma[ransub_inextp];
  if(mj<0) mj=mj+RANSUB_MAX;
  ransub_ma[ransub_inext]=mj;
  return(mj);
  }
 
double ransub () {
  long mj;
  ransub_inext++;
  if(ransub_inext==56) ransub_inext=1;
  ransub_inextp++;
  if(ransub_inextp==56) ransub_inextp=1;
  mj=ransub_ma[ransub_inext]-ransub_ma[ransub_inextp];
  if(mj<0) mj=mj+RANSUB_MAX;
  ransub_ma[ransub_inext]=mj;
  return((double)mj/RANSUB_MAX);
  }
 

