/* rndma = random matrix 
 *
 * MainSwitch options:
 * 0 generate a matrix and get its top eigval
 *       (or: the eigvals closest to lambda_0)
 *   AND repeat this for several random seeds
 *   OR: get number of small clusters (nosc) only
 * 100 ---REMOVED--- (ALMOST 0)
 *     generate a matrix and compute the product matrix of its rows:
 *     f(i,k)=sum_{j=1..n}[A_{ij}][A_{kj}]
 * 101 (ALMOST 100)
 *     get the average row product value for every bin defined on a
 *     log scale of i/j
 *     # of bins = # of vertices
 * 10,11,12 generating one matrix: 
 *     10: ER (Erdos-Renyi), 11: BA (Barabassi-Albert) or 12: UC (uncorrelated)
 * 2 read a matrix (in row list format) and get its top eigvals
 *       / the eigvals closest to lambda_0
 * 200 checking the Lanczos iteration:
 *     generate a matrix and get all eigvals of it using the NumRec method
 * 20 (SIMILAR to 0) 
 *    generate the matrix, store the eigvals lambda_1, lambda_2 and
 *    lambda_N the the largest row sums, store also the inverse
 *    participation ratios of the eigenvectors v_1, v_2 and v_N
 *    AND: repeat this for several system sizes and random seeds
 *    EXTRA: as a patch, using the BAT variable to give the size of the SW
 *    matrix in the different runs
 */

/******************************************************/

int ii,*RndSeed_main,RndSeed_main_index;

#include "rndma_lib.c"

int main( int NArg, char * ArgStr[] ) {

  _E("STARTED\t\t");system("date");
  _E("Initializing, please wait... \n"); 
  Start( NArg, ArgStr );
  _E("...finished.\n");


  switch(MainSwitch){
  case 0: case 101:{ 
          int seed_main;
	  

	  MS0_Init_Outside();
	  for(seed_main=MS0_FirstSeed;
	      seed_main<=MS0_LastSeed;
	      seed_main++
	     ){
	    
		  /* looking for largest eigval (Lanz_Fold=0)
		   * OR: for internal spectral density (Lanz_Fold=1)
		   */
                  MS0_Init_Inside();
		  MS0_Run( MainSwitch, seed_main, Lanz_Fold,
			   Lanz_Stretch,
			   &(MS0_NofIter[seed_main]), 
			   MS0_Lambda[seed_main],
			   MS0_Beta[seed_main], 
			   MS0_Error[seed_main], 
			   &(MS0_SumOmegaAbs[seed_main]),
			   MS0_Degree,
			   MS0_Lambda0num_I,
			   MS0_Lambda0_I,
			   MS0_NofIter_I[seed_main],
			   MS0_Lambda_I[seed_main],
			   MS0_Beta_I[seed_main],
			   MS0_Error_I[seed_main],
			   MS0_SumOmegaAbs_I[seed_main],
			   MS0_RowProd_Stat,
			   MS0_RowProd_StatCounter,
			   MS0_NOSC
			 );
		  MS0_Clean_Inside();
	  }


	  MS0_EvalSave_AtEnd( MainSwitch, Lanz_Fold, MS0_NofIter,
			      MS0_Lambda, 
			      MS0_Beta, MS0_Error, MS0_SumOmegaAbs,
			      MS0_Lambda0num_I,
			      MS0_Lambda0_I,
			      MS0_NofIter_I,
			      MS0_Lambda_I,
			      MS0_Beta_I,
			      MS0_Error_I,
			      MS0_SumOmegaAbs_I,
			      MS0_RowProd_Stat,
			      MS0_RowProd_StatCounter
			    );
	  break;
  }  
  case 10:{
          Model = 0;

          MS1_nVertex_num=6;
	  MS1_nVertex_arr=lvector(1,MS1_nVertex_num);
	  MS1_nVertex_arr[1]=300;
	  MS1_nVertex_arr[2]=1000;
	  MS1_nVertex_arr[3]=3000;
	  MS1_nVertex_arr[4]=10000;
	  MS1_nVertex_arr[5]=30000;
	  MS1_nVertex_arr[6]=100000;

	  MS1_nEdge_num=3;
	  MS1_nEdge_arr=lvector(1,MS1_nEdge_num);

	  for(ii_MS1_nV=1;ii_MS1_nV<=MS1_nVertex_num;ii_MS1_nV++){
	          ER_nVertex_mod10000 = MS1_nVertex_arr[ii_MS1_nV] % 10000;
	          ER_nVertex_div10000 = MS1_nVertex_arr[ii_MS1_nV] / 10000;
		  MS1_tmp_nV = (unsigned long)(  10000*ER_nVertex_div10000
					       + ER_nVertex_mod10000
					      );

		  MS1_nEdge_arr[1]=(int)floor(0.001*0.5*SQR(MS1_tmp_nV));
		  MS1_nEdge_arr[2]=(int)floor(0.003*0.5*SQR(MS1_tmp_nV));
		  MS1_nEdge_arr[3]=(int)floor(0.01*0.5*SQR(MS1_tmp_nV));

		  for(ii_MS1_nE=1;ii_MS1_nE<=MS1_nEdge_num;ii_MS1_nE++){
		         ER_nEdge_mod10000 = MS1_nEdge_arr[ii_MS1_nE] % 10000;
			 ER_nEdge_div10000 = MS1_nEdge_arr[ii_MS1_nE] / 10000;
		  
			 MS1_Init();
			 MS1_Run();
			 MS1_Save();
		  }
	  }
	  break;
  }
  case 11:{
          Model = 1;

	  MS1_T_num=6;
	  MS1_T_arr=lvector(1,MS1_T_num);
	  MS1_T_arr[1]=300;
	  MS1_T_arr[2]=1000;
	  MS1_T_arr[3]=3000;
	  MS1_T_arr[4]=10000;
	  MS1_T_arr[5]=30000;
	  MS1_T_arr[6]=100000;

	  for(ii_MS1_T=1;ii_MS1_T<=MS1_T_num;ii_MS1_T++){
	          BA_Tmod10000 = MS1_T_arr[ii_MS1_T] % 10000;
	          BA_Tdiv10000 = MS1_T_arr[ii_MS1_T] / 10000;
		  
	          MS1_Init();
		  MS1_Run();
		  MS1_Save();
	  }
	  break;
  }
  case 12:{
          Model = 2;

          MS1_nVertex_num=1;
	  MS1_nVertex_arr=lvector(1,MS1_nVertex_num);
	  MS1_nVertex_arr[1]=3000;
	  /*          MS1_nVertex_num=7;
	   *	  MS1_nVertex_arr=lvector(1,MS1_nVertex_num);
	   *	  MS1_nVertex_arr[1]=300000;
	   *	  MS1_nVertex_arr[2]=100000;
	   *	  MS1_nVertex_arr[3]=30000;
	   *	  MS1_nVertex_arr[4]=10000;
	   *	  MS1_nVertex_arr[5]=3000;
	   *	  MS1_nVertex_arr[6]=1000;
	   *	  MS1_nVertex_arr[7]=300;
	   */

	  MS1_Gamma_num=3;
	  MS1_Gamma_arr=dvector(1,MS1_Gamma_num);
	  MS1_Gamma_arr[1]=2.5;
	  MS1_Gamma_arr[2]=3.0;
	  MS1_Gamma_arr[3]=4.0;

	  for(ii_MS1_nV=1;ii_MS1_nV<=MS1_nVertex_num;ii_MS1_nV++){
	          UC_nVertex_mod10000 = MS1_nVertex_arr[ii_MS1_nV] % 10000;
	          UC_nVertex_div10000 = MS1_nVertex_arr[ii_MS1_nV] / 10000;

		  for(ii_MS1_G=1;ii_MS1_G<=MS1_Gamma_num;ii_MS1_G++){
		          UC_Gamma = MS1_Gamma_arr[ii_MS1_G];
			  		  
			  MS1_Init();
			  MS1_Run();
			  MS1_Save();
		  }
	  }
	  break;
  }
  case 2:{ 
	  MS2_InitAndReadMatrix();
	  MS2_Run();
	  MS2_EvalAndSave();
          break;
  }
  case 200:{ 
	  MS200_Init();
	  MS200_Run( RndSeed, MS200_Degree );
	  MS200_EvalAndSave();
          break;
  }
  case 20:{ 

	  MS20_Init_Outside();
	  for(MS20_BAT_index=1;
	      MS20_BAT_index<=MS20_Nof_BAT;
	      MS20_BAT_index++){

	      for(MS20_seed_main=MS20_FirstSeed;
		  MS20_seed_main<=MS20_LastSeed;
		  MS20_seed_main++
		  ){
	    
	              MS20_Init_Inside( MS20_BAT_index );
	              /* looking for largest eigval */
		      MS20_Run( MS20_seed_main,
				MS20_BAT_index,
				MS20_Lambda1,
				MS20_Lambda2,
				MS20_LambdaN,
				MS20_Degree,
				MS20_LD[MS20_seed_main][MS20_BAT_index],
				&MS20_IPR_1, &MS20_IPR_2, &MS20_IPR_N
				);
		      MS20_EvalSave_Inside(
			    MS20_seed_main,
			    MS20_BAT_index,	   
			    MS20_Lambda1,
			    MS20_Lambda2,
			    MS20_LambdaN,
			    MS20_LD,
			    MS20_IPR_1, MS20_IPR_2, MS20_IPR_N
			    );
                      MS20_Clean_Inside();
	    }
	  }


	  MS20_EvalSave_AtEnd( MS20_Lambda1, 
			       MS20_Lambda2, 
			       MS20_LambdaN, 
			       MS20_LD
			       );
	  break;
  }  
  default: {
          fprintf(stderr,"ERROR: no such MainSwitch: %d\n",MainSwitch);
	  _E("Exiting to system.\n");
	  exit(-1);
  }
  }

  _E("FINISHED\t\t");system("date");
}
