/* default MainSwitch and input file name */
int MainSwitch, MainSwitch_DEFAULT = 0; 
char IFN[100], *IFN_DEFAULT = "rndma.par";

/**************************  global constants  ********************/  

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "nrutil.h"
#include "nrutil_ext_v_2000_08_23.c"
#include "base_v_2000_06_07.c"
#include "readpar_v_2000_08_11.c"
#include "tred2_d_ul.c"
#include "tqli_il_d_ul.c"
#include "indexx_ov_d_ul.c"
#include "ran_Ellak.c"
#include "d_sort.c"
#include "gauss_ranksz.c"
#include "pythag.c"
#include "tred2_ne.c"
#include "tqli_ne_il.c"
#include "sort.c"
#include "indexx_ul.c"

#define MY_EXIT {_E("Exiting to system.\n");exit(-1);}
#define MY_STRLEN 300

/* macros to maintain type and precision of variables
 * the macros of nrutil.h all use float precision
 */
#ifndef MY_ABS
#define	MY_ABS(a)	 ( (a) >= 0.0 ? (a) : (-1.0)*(a) )
#endif
#ifndef MY_SQR
#define MY_SQR(a)               ( (a)*(a) )
#endif
#ifndef MY_MAX
#define MY_MAX(a,b)     ( (a) >= (b) ? (a) : (b) )
#endif
#ifndef MY_MIN
#define MY_MIN(a,b)     ( (a) >= (b) ? (b) : (a) )
#endif

/*********************  global parameters  *****************************/

static int RndSeed, NofSeeds, Model, 
  ER_nVertex_mod10000, ER_nVertex_div10000,
  ER_nEdge_mod10000, ER_nEdge_div10000,
  BA_M0, BA_M, BA_Tmod10000, BA_Tdiv10000, 
  UC_nVertex_mod10000, UC_nVertex_div10000, UC_minDegree,
  UC_NTryLimit_mod10000, UC_NTryLimit_div10000, 
  SW_nVertex_mod10000, SW_nVertex_div10000, SW_k,
  Lanz_DebugSwitch, Lanz_M, Lanz_Nd, Lanz_NIterMin, tqli_ItMax,
  Lanz_Fold, Lanz_Stretch, Lanz_NOSC, ARP_NofBins;
static double UC_Gamma, UC_pN, SW_p, Lanz_MEps, Lanz_REps, Lanz_Phi,
  Lanz_IW, Lanz_TeGS; 
static char OFN[MY_STRLEN], Lanz_DebugName[MY_STRLEN];

int* IPar[] = {&RndSeed, &NofSeeds, &Model, 
	       &ER_nVertex_mod10000, &ER_nVertex_div10000,
	       &ER_nEdge_mod10000, &ER_nEdge_div10000,
	       &BA_M0, &BA_M, &BA_Tmod10000, &BA_Tdiv10000, 
	       &UC_nVertex_mod10000, &UC_nVertex_div10000, &UC_minDegree,
	       &UC_NTryLimit_mod10000, &UC_NTryLimit_div10000, 
	       &SW_nVertex_mod10000, &SW_nVertex_div10000, &SW_k,
	       &Lanz_DebugSwitch,
	       &Lanz_M, &Lanz_Nd, &Lanz_NIterMin, &tqli_ItMax,
	       &Lanz_Fold, &Lanz_Stretch, &Lanz_NOSC, &ARP_NofBins};
char* IParName[] = {"RndSeed", "NofSeeds", "Model", 
		    "ER_nVertex_mod10000", "ER_nVertex_div10000", 
		    "ER_nEdge_mod10000", "ER_nEdge_div10000", 
		    "BA_M0", "BA_M", "BA_Tmod10000", "BA_Tdiv10000",  
		    "UC_nVertex_mod10000", "UC_nVertex_div10000",
		    "UC_minDegree", 
		    "UC_NTryLimit_mod10000", "UC_NTryLimit_div10000", 
		    "SW_nVertex_mod10000", "SW_nVertex_div10000", "SW_k",
		    "Lanz_DebugSwitch",
		    "Lanz_M", "Lanz_Nd", "Lanz_NIterMin",
		    "tqli_ItMax", "Lanz_Fold", "Lanz_Stretch",
		    "Lanz_NOSC", "ARP_NofBins"}; 
double* DPar[] = {&UC_Gamma, &UC_pN, &SW_p, &Lanz_MEps, &Lanz_REps,
		  &Lanz_Phi, &Lanz_IW, &Lanz_TeGS}; 
char* DParName[] = {"UC_Gamma", "UC_pN", "SW_p", "Lanz_MEps", "Lanz_REps",
		    "Lanz_Phi", "Lanz_IW", "Lanz_TeGS"};
char *SPar[] = {OFN, Lanz_DebugName};
char *SParName[] = {"OFN", "Lanz_DebugName"};


int IParNum = sizeof(IPar)/sizeof(int*),
  DParNum = sizeof(DPar)/sizeof(double*),
  SParNum = sizeof(SPar)/sizeof(char*);

/************************  global variables  ***************************/

unsigned long **NEI, BA_T, UC_NTryLimit,*MS0_NOSC;

/*
 * NEI[i][j] (neighbor index) is the index of the jth neighbor of the
 *           ith vertex  
 *
 * NOSC: number of small clusters
 * nosc[1]: # of clusters with 1 vertex (isolated point)
 * nosc[2]: # of clusters with 2 vertices (isolated pair)
 * nosc[3]: # of clusters with 3 vertices and 2 edges 
 * nosc[4]: # of clusters with 3 vertices and 3 edges 
 */


/* variables of MainSwitch=0 for LanzFold=0, ie. top eigval */
int *MS0_NofIter,MS0_FirstSeed,MS0_LastSeed;
unsigned long *MS0_Degree,MS0_nVertex,MS0_nEdge,MS0_UC_nEdge;
double **MS0_Lambda, **MS0_Beta, **MS0_Error, *MS0_SumOmegaAbs,
  *MS0_RowProd_Stat,*MS0_RowProd_StatCounter;

/* variables of MainSwitch=0 for LanzFold=1, ie. internal (I) eigvals */
int MS0_Lambda0num_I,**MS0_NofIter_I;
double *MS0_Lambda0_I,***MS0_Lambda_I,***MS0_Beta_I,***MS0_Error_I,
  ***MS0_SumOmegaAbs_I;


int ii_MS1_T, MS1_T_num, ii_MS1_nE, MS1_nEdge_num, ii_MS1_nV,
  MS1_nVertex_num, ii_MS1_G, MS1_Gamma_num;
unsigned long *MS1_T_arr, *MS1_nEdge_arr, *MS1_nVertex_arr,
  MS1_tmp_nV, MS1_nVertex, MS1_nEdge, *MS1_Degree, MS1_UC_nEdge;
double *MS1_Gamma_arr;

unsigned long MS2_nVertex, *MS2_Degree;

unsigned long MS200_nVertex, MS200_nEdge, *MS200_Degree;
float *MS200_Lambda;



int MS20_seed_main,MS20_Nof_BAT,MS20_BAT_index,
  MS20_FirstSeed,MS20_LastSeed,MS20_NofLD;
unsigned long MS20_nVertex, MS20_nEdge, ***MS20_LD, *MS20_Degree,
  MS20_UC_nEdge, *MS20_BAT;
double **MS20_Lambda1,**MS20_Lambda2,**MS20_LambdaN,
  MS20_IPR_1, MS20_IPR_2, MS20_IPR_N;
FILE *MS20_OFP;

/*
 * MS20_Nof_BAT: the number of T values (BA model) 
 * to be checked 
 * MS20_BAT[i] the ith T value to be checked
 * MS20_BAT_index: indexes the array MS20_BAT 
 * MS20_LD[seed][BAT_index]: 
 *      array holding the MS20_NofLD largest degrees of the graph
 *      generated with seed and the given BA_t value
 * MS20_Lambda1[seed][BAT_index] is the largest eigval when using this
 *      rnd seed and the system size indexed by BAT_index
 * MS20_IPR_#: inverse participation ratios of the eigenvector v_#
 */

/**********************   prototypes   **********************/

void Start( int narg, char *argstr[] );

void GenerateMatrix( int seed, int model, 
		     unsigned long er_nvertex, unsigned long er_nedge, 
		     unsigned long ba_m0, unsigned long ba_m, 
		     unsigned long ba_t, 
		     double uc_gamma, unsigned long uc_nvertex, 
		     unsigned long *uc_nedge,
		     unsigned long uc_mindegree, double uc_pN,
		     unsigned long uc_ntrylimit,
		     unsigned long sw_nvertex,
		     int sw_k,
		     double sw_p,
		     unsigned long *degree
		   );
void GenerateMatrix_ER( int seed, unsigned long er_nvertex, 
			unsigned long er_nedge, unsigned long *degree
                      );
void GenerateMatrix_BA( int seed, unsigned long ba_m0, 
			unsigned long ba_m, unsigned long ba_t,
			unsigned long *degree 
       		      ); 
void GenerateMatrix_UC( int seed, int model,
			unsigned long uc_gamma, 
			unsigned long uc_nvertex,
			unsigned long *uc_nedge,
			unsigned long uc_mindegree,
			double uc_pN, unsigned long uc_ntrylimit,
			unsigned long *degree
                      );
void GenerateMatrix_SW( int seed, int model,
			unsigned long sw_nvertex,
			int sw_k,
			double sw_p,
			unsigned long *degree
                      );
void GetEigVals( unsigned long n, int fold_sw, 
		 double spec_shift, int spec_stretch,
		 double folding_point, unsigned long *degree,
		 int *n_of_iter,
		 double *return_lambda, double *return_beta, 
		 double *return_error, double *return_sum_omega_abs,
		 double *ipr
	       );

void SortedEigVals_NumRec( float **a, int n, float *eigval );

/*------------------------*/

void MS0_Init_Outside();
void MS0_Init_Inside();
void MS0_Run( int main_switch, int seed, int fold_switch, int stretch,
	      int *n_of_iter, double *lambda,
	      double *beta, double *error, double *sum_omega_abs, 
	      unsigned long *degree,
	      int lambda_0_num_I, double *lambda_0_I,
	      int *n_of_iter_I, double **lambda_I,
	      double **beta_I, double **error_I, double **sum_omega_abs_I,
	      double *rowprod_stat, double *rowprod_statcounter,
	      unsigned long *nosc	      
	    );
void MS0_Clean_Inside();
void MS0_EvalSave_AtEnd( int main_switch, int fold, int *n_of_iter,
			 double **lambda, 
			 double **beta, double **error, 
			 double *sum_omega_abs,
			 int lambda_0_num_I, double *lambda_0_I,
			 int **n_of_iter_I, double ***lambda_I,
			 double ***beta_I, double ***error_I,
			 double ***sum_omega_abs_I,
			 double *rowprod_stat, double *rowprod_statcounter
		       );
void MS0_Cnosc( unsigned long *nosc );

void MS1_Init();
void MS1_Run();
void MS1_Save();

void MS2_InitAndReadMatrix();
void MS2_Run();
void MS2_EvalAndSave();

void MS200_Init();
void MS200_Run( int seed, unsigned long *degree );
void MS200_EvalAndSave();

void MS20_Init_Outside();
void MS20_Init_Inside( int bat_index );
void MS20_Run( int seed, int bat_index, 
	       double **lambda1, 
	       double **lambda2, 
	       double **lambdaN, 
	       unsigned long *degree,
	       unsigned long *ld,
	       double *ipr1, double *ipr2, double *iprN
	     );
void MS20_Clean_Inside();
void MS20_EvalSave_Inside( int seed, int bat_index, 
			   double **lambda1, 
			   double **lambda2, 
			   double **lambdaN, 
			   unsigned long ***ld,
			   double ipr1, double ipr2, double iprN
			   );
void MS20_EvalSave_AtEnd( double **lambda1, 
			  double **lambda2, 
			  double **lambdaN, 
			  unsigned long ***ld 
			  );

/***************** definitions *************************/

void Start( int narg, char *argstr[] ){

  /* 0 */
  switch(narg){
  case 1: { 
          MainSwitch = MainSwitch_DEFAULT; 
          strcpy(IFN,IFN_DEFAULT);
          fprintf(stderr,"(using default: MainSwitch = %d, input file name = %s)\n"
,MainSwitch_DEFAULT,IFN_DEFAULT);fflush(stderr);
          break; 
  }
  case 3: { 
          MainSwitch = atoi(argstr[1]); 
          strcpy(IFN,argstr[2]);
          break; 
  }
  default:{
          _E("Usage:  either \"rndma <MainSwitch> <input file name>\",");
          _E("        or     \"rndma\" .\n");
          MY_EXIT;
  }}


  /* 1 */
  readpar( "start", IFN, 
	   IPar, IParName, IParNum, 
	   DPar, DParName, DParNum, 
	   SPar, SParName, SParNum
	 );



  /* 2 */
  /* parameter check, emergency exit */
  if(Model==20){
          fprintf(stderr,"Sorry, Model=20 not implemented yet.\n");
	  MY_EXIT;
  }
  if((Lanz_Fold==10)&&(MainSwitch!=0)){
    fprintf(stderr,"Sorry, Lanz_Fold=10 implemented only for MainSwitch=0\n");
    MY_EXIT;
  }
  if((Model==3)&&((SW_k<1)||(SW_k%2==1))){
    fprintf(stderr,"Sorry, SW_k has to be an even (positive) number).\n");
    MY_EXIT;
  }
  if((MainSwitch==20)&&(Lanz_Nd<2)){
    fprintf(stderr,"Sorry, MainSwitch=20 needs Lanz_Nd >= 2\n");
    MY_EXIT;
  }

}

/**************************************/

void GenerateMatrix( int seed, int model, 
		     unsigned long er_nvertex, unsigned long er_nedge, 
		     unsigned long ba_m0, unsigned long ba_m, 
		     unsigned long ba_t, 
		     double uc_gamma, unsigned long uc_nvertex, 
		     unsigned long *uc_nedge,
		     unsigned long uc_mindegree, double uc_pN,
		     unsigned long uc_ntrylimit,
		     unsigned long sw_nvertex,
		     int sw_k,
		     double sw_p,
		     unsigned long *degree
                   ){


  switch(model){
  case 0: default: {
          GenerateMatrix_ER( seed, er_nvertex, er_nedge, degree );
	  break;
  }
  case 1:{
          GenerateMatrix_BA( seed, ba_m0, ba_m, ba_t, degree );
	  break;
  }
  case 2: case 20:{
          GenerateMatrix_UC( seed, model, 
			     uc_gamma, uc_nvertex, uc_nedge, uc_mindegree,
			     uc_pN, uc_ntrylimit, degree
			    ); 
	  break;
  }
  case 3:{
          GenerateMatrix_SW( seed, model, sw_nvertex, sw_k, sw_p, degree );
	  break;
  }
  }
}

/*--------------------------------------------*/

void GenerateMatrix_ER( int seed, unsigned long er_nvertex, 
			unsigned long er_nedge, unsigned long *degree
		      ){
  /* generating the adjacency matrix of an ER graph */

  unsigned long i,*pam,j,tmp_l1,tmp_l2,*nei_ncl,*nei_nch,
    *c_lnl,ie,iv;
  int connected;

  /* 
   * (pam[i],pam[er_nedge+i]) are the vertices connected by the ith edge,
   *                   pam = pair member
   * c_lnl: counter for the length of the neighbor list
   *        c_lnl[i]-1 is the number of neighbors of vertex i
   * nei_ncl[i] is the starting(l=low) index of the ith row of NEI[][]
   *     nch is the last(h=high)
   */



  /* rnd seed, memory allocation */
  seedksz(seed);
  pam=lvector(1,2*er_nedge);
  c_lnl=lvector( 1, er_nvertex );
  nei_ncl=lvector( 1, er_nvertex );
  nei_nch=lvector( 1, er_nvertex );



  /* 1.
   * pair lists
   */

  /*_E("ER_1\t");system("date");*/
  for(iv=1;iv<=er_nvertex;iv++){ degree[iv] = 0; }
  for(ie=1;ie<=er_nedge;ie++){

          /*if(ie%10000==0){fprintf(stderr,"ie=%ld\t",ie);system("date");}*/
          do{ 
	          tmp_l1 = 1+(int)floor((double)er_nvertex*ranksz());
	          tmp_l2 = 1+(int)floor((double)er_nvertex*ranksz());
	  }while(tmp_l1==tmp_l2);


	  j = 1;
	  connected = 0;
	  while(j<ie){
	          if(  ((pam[j]==tmp_l1)&&(pam[er_nedge+j]==tmp_l2))
		     ||((pam[j]==tmp_l2)&&(pam[er_nedge+j]==tmp_l1))
		    ){
		          connected = 1; /* this pair of vertices has 
					  * been connected already
					  */
			  j = ie; /* exit "while" loop */
		  }
		  else{
		          j++;
		  }
	  }

	  /* if chosen pair is not connected, connect them */
	  if(connected==0){
	          pam[ie] = tmp_l1;
	          pam[er_nedge+ie] = tmp_l2;
		  degree[tmp_l1]++;
		  degree[tmp_l2]++;
	  }
	  /* if chosen pair is already connected, try again */
	  else{
	          ie--;
	  }
  }



  /* 2. 
   * transforming the adjacency matrix from the previous pam
   * into row list format:
   * the non-zero entries (all of them being 1) of row i 
   * will be listed in the array NEI[i]
   */
  
  /* 2.1 init */
  for(iv=1;iv<=er_nvertex;iv++){
          nei_ncl[iv]=1;
	  if(degree[iv]==0){ nei_nch[iv] = 1; }
	  else{ nei_nch[iv] = degree[iv]; }
  }
  NEI=lmatrix_vrl(1,er_nvertex,nei_ncl,nei_nch);


  /* 2.2 creating neighbor lists */
  for(iv=1;iv<=er_nvertex;iv++){
	  c_lnl[iv]=1;
  }
  for(ie=1;ie<=er_nedge;ie++){
          i = pam[ie];
          j = pam[er_nedge+ie];
          NEI[i][c_lnl[i]] = j;
	  c_lnl[i]++;
          NEI[j][c_lnl[j]] = i;
	  c_lnl[j]++;
  }


  /* 2.3
   * counting small clusters
   */
  
  





  free_lvector(pam,1,2*er_nedge);
  free_lvector( c_lnl, 1, er_nvertex );
  free_lvector( nei_ncl, 1, er_nvertex );
  free_lvector( nei_nch, 1, er_nvertex );
}

/*----------------------------------*/

void GenerateMatrix_BA( int seed, unsigned long ba_m0, 
			unsigned long ba_m, unsigned long ba_t,
			unsigned long *degree 
		      ){
  /* generating the adjacency matrix of a BA graph */

  unsigned long i,j,k,kk,*tea,sod,*c_lnl,*nei_ncl,*nei_nch;
  double tmpd,*p,one_over_sod;
  int connected;

  /* tea[m*(i-m0-1)+j] (table of edges' arriving ends)
   *    is the index of the vertex to which the 
   *    (m0+i)th vertex is connected via its
   *    jth edge (out of those m which it brings to the graph when it
   *    is attached)
   *    tea is the index of the vertex where this new edge arrives
   * sod: sum of degrees
   * p: the length of (p[i-1],p[i]) is degree[i]/sum_j(degree[j]) 
   * c_lnl[i]: counter for the length of the neighbor list of vertex i
   * nei_ncl[i] is the starting index of the ith row of nei[][]
   *     nch           last
   */


  /* rnd seed, mem. alloc. */
  seedksz(seed);
  for(i=1;i<=ba_m0+ba_t;i++){ degree[i]=0; }
  tea=lvector( 1, ba_m*ba_t );
  c_lnl=lvector( 1, ba_m0+ba_t );
  nei_ncl=lvector( 1, ba_m0+ba_t );
  nei_nch=lvector( 1, ba_m0+ba_t );



/*fprintf(stderr,"BA_1\t");
 *system("date");
 */

  /* 1.
   * adding new vertices 
   */
  sod=0;
  for(i=ba_m0+1;i<=ba_m0+ba_t;i++){

if(i%10000==0){
fprintf(stderr,"i=%ld\t",i);
system("date");
}

      /* 1.1
       * weights 
       */
      p=dvector(0,i-1);
	  
      /* when attaching first new vertex */
      if(i==ba_m0+1){
	  for(j=0;j<=ba_m0;j++){
	      p[j]=(double)j/(double)ba_m0;
	  }
      }
      /* at later steps */
      else{
	  p[0]=0.0;
	  one_over_sod = 1.0/(double)sod;
	  for(j=1;j<i;j++){ 
	      p[j] = p[j-1] + (double)degree[j]*one_over_sod; 
	  }
      }


      /* 1.2
       * adding to the book-keeping the m edges leaving from the new vertex 
       */
      for(k=1;k<=ba_m;k++){
	  do{
	      /* pick an older vertex randomly using the weights
	       * generated above 
	       */
	      tmpd = ranksz();
	      j=1;
	      while(p[j]<tmpd){j++;}

	      /* check if the new vertex is already connected to the
	       * chosen old one
	       */
	      connected=0;
	      kk=1;
	      while(kk<k){
		  if(tea[ba_m*(i-ba_m0-1)+kk]==(unsigned long)j){ 
		      connected=1; /* set flag */
		      kk=k-1; /* exit check loop */
		  }
		  kk++;
	      }
	  }while(connected==1);

	  /* if not yet, connect */
	  tea[ba_m*(i-ba_m0-1)+k]=j;
	  degree[i]++;
	  degree[j]++;
	  sod += 2;
      }

      free_dvector(p,0,i-1);
  }



  /* 2.
   * transforming the adjacency matrix from the previous tea format
   * into row list format: 
   * the non-zero entries (all of them are 1) of row i are stored in
   * the array NEI[i]
   */


  /* init */
  for(i=1;i<=ba_m0+ba_t;i++){
      nei_ncl[i]=1;
      nei_nch[i]=degree[i]; /* is >= 1 because of the generating
			     * method of the BA matrix 
			     */
      c_lnl[i]=1;
  }
  NEI=lmatrix_vrl(1,ba_m0+ba_t,nei_ncl,nei_nch);



  /* 2.1 creating neighbor lists */

  /* 2.1/a 
   * inserting the entries of the upper triangle into the lists
   * (ie. inserting every edge into the list of the vertex where it
   * starts from)
   */
  for(i=ba_m0+1;i<=ba_m0+ba_t;i++){
      for(k=1;k<=ba_m;k++){
	  NEI[i][c_lnl[i]]=tea[ba_m*(i-ba_m0-1)+k];
	  c_lnl[i]++;
      }
  }
  /* 2.1/b
   * inserting every edge into the list of the vertex of its arrival 
   */
  for(j=ba_m0+1;j<=ba_m0+ba_t;j++){
      for(k=1;k<=ba_m;k++){
	  i=tea[ba_m*(j-ba_m0-1)+k];
	  NEI[i][c_lnl[i]]=j;
	  c_lnl[i]++;
      }
  }



  free_lvector( tea, 1, ba_m*ba_t );
  free_lvector( c_lnl, 1, ba_m0+ba_t );
  free_lvector( nei_ncl, 1, ba_m0+ba_t );
  free_lvector( nei_nch, 1, ba_m0+ba_t );
}

/*-------------------------------------*/

void GenerateMatrix_UC( int seed, int model,
			unsigned long uc_gamma, 
			unsigned long uc_nvertex,
			unsigned long *uc_nedge,
			unsigned long uc_mindegree,
			double uc_pN, unsigned long uc_ntrylimit,
			unsigned long *degree
                      ){

  int connected,remaining_graph_is_complete,exit_flag,
    connected_to_jv1,connected_to_lv_iv2;
  unsigned long kmin,kmax,l,k,i,*lv,iv,*nei_ncl,*nei_nch,*c_lnl,j,
    nov,novR,iv1,iv2,tmp_iv1,tmp_lv,jv1,jv2,tmp_counter,
    index_of_jv2_on_jv1s_list,index_of_jv1_on_jv2s_list;
  double tmpd1,tmpd2,*p,spk,tmpd;


  /* nov: the total number of valencies
   * novR: nov remaining
   * lv: valency list
   * iv1: index of valency
   * kmin: minimum degree of a vertex
   * p: the length of the (p[k-1],p[k]) interval is 
   *    proportional to k^{-uc_gamma}
   * spk: sum for normalizing p[k]
   * c_lnl: counter for length of neighbor list
   *        c_lnl[i]-1 is current number of the neighbors of vertex i
   */


fprintf(stderr,"Sorry, UC not finished yet\n");
fprintf(stderr,"Probably the routine GenerateMatrix_UC gets stuck somewhere\n");
fprintf(stderr,"It may be ok now -- after some corrections\n");
MY_EXIT;


  /*
   * generating the adjacency matrix of a UC graph:
   * every point has
   * a fixed degree distribution ~ k^{-gamma}, 
   * but the degrees of any two vertices are uncorrelated (hopefully)
   *
   * kmin = uc_mindegree
   * kmax = (if model=2) uc_nvertex
   *        (if model=20) such that pN = uc_pN
   */




  /* 1. 
   * DEGREE SEQUENCE 
   * fixed, uncorrelated: P(k) ~ k^{-uc_gamma}
   */

  /* kmin and kmax */
  kmin=uc_mindegree;
  switch(model){
  case 2: default:{ kmax=uc_nvertex-1; break; }
  case 20:{
          /* <k> =   sum_{kmin,kmax} l^(1-gamma) 
	   *       / sum_{kmin,kmax} l^{-gamma}
	   * chosing the first such kmax, for which the rhs. is >= uc_pN
	   */
          if( uc_pN < kmin ){
	          fprintf(stderr,"rndma ERROR: UC_pN < UC_minDegree\n");
		  MY_EXIT;
	  }
          tmpd1=tmpd2=0.0;
	  l=kmin;
	  while(    ( l == kmin )
		 || (    ( l > kmin ) 
		      && ( l <= uc_nvertex-1 )
		      && ( tmpd1 < tmpd2*uc_pN ) 
		    )
	       ){
	          tmpd1+=pow((double)l,1.0-uc_gamma);
	          tmpd2+=pow((double)l,-uc_gamma);
		  kmax=l;
		  l++;
	  }
	  if(l==uc_nvertex){
	          fprintf(stderr,"rndma WARNING: l has reached uc_nvertex\n");
		  fprintf(stderr,"               pN = UC_pN might not hold\n");
	  }

fprintf(stderr,"kmin=%ld\tkmax=%ld\n",kmin,kmax);fflush(stderr);
          break;
  }
  }

  /* seed, aux. arrays */
  seedksz(seed);
  p=dvector(kmin-1,kmax);
  p[kmin-1]=0.0;
  spk=0.0;
  for(k=kmin;k<=kmax;k++){ 
          spk += pow(k,-uc_gamma);
	  p[k] = spk;
  }

  /* degrees of vertices */
  nov=0;
  for(i=1;i<=uc_nvertex;i++){
          tmpd=(double)spk*ranksz();
	  k=kmin-1;
	  while(p[k]<tmpd){ k++; }
	  degree[i]=k;
	  nov+=k;
  }
  free_dvector(p,kmin-1,kmax);

  /* the sum of valencies has to be an even number,
   * if it's not, adding an extra valency to one vertex 
   */
  if((nov%2)==1){
          i=1;
	  while(degree[i]==uc_nvertex-1){ i++; }
          degree[i]++;
	  nov++;
  }

  /* mem. alloc. to NEI (neighbor index list)
   * init NEI
   */
  nei_ncl=lvector(1,uc_nvertex);
  nei_nch=lvector(1,uc_nvertex);
  for(i=1;i<=uc_nvertex;i++){
          nei_ncl[i]=1;
	  nei_nch[i]=degree[i];
  }
  NEI=lmatrix_vrl(1,uc_nvertex,nei_ncl,nei_nch);
  free_lvector(nei_ncl,1,uc_nvertex);
  free_lvector(nei_nch,1,uc_nvertex);
  for(i=1;i<=uc_nvertex;i++){
          for(j=1;j<=degree[i];j++){
	          NEI[i][j] = 0;
	  }
  }

  /* initializing valency list */
  lv=lvector(1,nov);
  iv=1;
  for(i=1;i<=uc_nvertex;i++){
          for(k=1;k<=degree[i];k++){
	          lv[iv]=i;
		  iv++;
	  }
  }

  /* initializing c_lnl */
  c_lnl=lvector(1,uc_nvertex);
  for(i=1;i<=uc_nvertex;i++){
          c_lnl[i] = 1;
  }





  
  /* 2. CONNECTING:
   *
   * decide whether remaining graph is already complete (fully connected)
   *
   * IF NO:
   * pick two valencies randomly, 
   * make sure they're not on the same vertex,
   * connect them: (i) connect the two vertices 
   *               (ii) remove the two valencies from the val. list
   *
   * IF YES:
   * pick two valencies (iv1 and iv2) from two different vertices of
   * the remaining graph and two vertices (jv1 and jv2) different from
   * each other and both lv[iv1] and lv[iv2]
   * SUCH that:     lv[iv1] and jv1 are not connected yet
   *            AND lv[iv2] and jv2 are not connected yet
   *            AND jv1 and jv2 are connected
   *
   * disconnect jv1 and jv2
   * connect lv[iv1] and jv1
   * connect lv[iv2] and jv2
   * remove the valencies iv1 and iv2 from the valency list
   *
   *
   * SPECIAL EXIT: 
   * if all remaining valencies are on one vertex:
   * stop the algorithm, it's only one vertex, will not harm
   */

  (*uc_nedge) = 0;
  novR = nov;
  while(    ( novR > 0 ) 
	 && ( lv[1] < lv[novR] )        
       ){
  
	  /* check if remaining graph is complete */
	  remaining_graph_is_complete = 1;
	  iv1=1;
	  while(iv1<=novR){
	          /* look for first valency on the list that is not
		   * on the vertex lv[iv1]
		   */
	          iv2=iv1+1;
		  while(  (iv2<=novR)
			&&(lv[iv2]==lv[iv1])
		    ){
		          iv2++;
		  }
		  /* if there is such a valency, proceed */
		  while(iv2<=novR){

			  /* deciding whether the vertices lv[iv1] and
			   * lv[iv2] are connected 
			   */
			  connected = 0;
		          j=1;
			  while(j<=degree[lv[iv1]]){
			          if(NEI[lv[iv1]][j]==lv[iv2]){
				          /* lv[iv1] and lv[iv2] are
					   * connected :
					   * exit j loop
					   */
				          connected = 1;
					  j = degree[lv[iv1]]+1;
				  }					  
				  else{
				          j++;
				  }
			  }
			  if(connected==0){
			          /* found a pair of vertices not
				   * connected: lv[iv1] and lv[iv2]
				   * exit iv1 and iv2 loops 
				   */
			          remaining_graph_is_complete = 0;
				  iv2 = novR+1;
				  iv1 = novR+1;
			  }     
			  else{
			          /* look for first valency on the
				   * list that is not on the vertex lv[iv2]
				   */
			          tmp_lv = lv[iv2];
				  while(lv[iv2]==tmp_lv){
				          iv2++; 
				  }
			  }

			  /* look for first valency on the
			   * list that is not on the vertex lv[iv1]
			   */
			  tmp_lv = lv[iv1];
				  while(lv[iv1]==tmp_lv){
				          iv1++; 
				  }
		  }
	  }


	  switch(remaining_graph_is_complete){
	  case 0: default:{ /* if remaining graph is not complete */

		  do{ 
		          /* pick two valencies from two different 
			   * vertices 
			   */
		          do{
			          iv1=1+(int)floor((double)novR*ranksz());
				  iv2=1+(int)floor((double)novR*ranksz());
			  }while(lv[iv1]==lv[iv2]);

			  /* see, whether their vertices are already
			   * connected 
			   */
			  connected = 0;
			  /* the degree of any vertex is >= kmin >= 1 */
			  j=1;
			  while(j<=degree[lv[iv1]]){
			          if(NEI[lv[iv1]][j]==lv[iv2]){
				          connected = 1;
				  }
				  else{
				          j++;
				  }
			  }

		  }while(connected==1);


		  /* connect the two vertices 
		   * and update book-keeping 
		   */
		  NEI[lv[iv1]][c_lnl[lv[iv1]]] = lv[iv2];
		  c_lnl[lv[iv1]]++;
		  NEI[lv[iv2]][c_lnl[lv[iv2]]] = lv[iv1];
		  c_lnl[lv[iv2]]++;

		  tmp_iv1 = iv1;
		  iv1 = (int)MIN(iv1,iv2);
		  iv2 = (int)MAX(tmp_iv1,iv2);
		  for(iv=iv2;iv<=novR-1;iv++){ lv[iv]=lv[iv+1]; }
		  novR--;
		  for(iv=iv1;iv<=novR-1;iv++){ lv[iv]=lv[iv+1]; }
		  novR--;

		  (*uc_nedge)++;

		  break;
	  }
	  case 1:{ /* if remaining graph is complete */

	          do{
		          exit_flag = 1;

		          /* first two vertices */
		          do{ 
			          iv1=1+(int)floor((double)novR*ranksz());
				  iv2=1+(int)floor((double)novR*ranksz());
			  }while(lv[iv1]==lv[iv2]);


			  /* first vertex of second pair */
			  tmp_counter = 0;
			  do{
			          /* pick a vertex differing from the
				   * previous two 
				   */
			          do{
				          jv1=1+(int)floor((double)uc_nvertex*ranksz());
					  tmp_counter++;
				  }while(  (tmp_counter<uc_ntrylimit)
					 &&(  (lv[iv1]==jv1)  
					    ||(lv[iv2]==jv1)  
					   )
					);   


				  /* if found such a vertex: */
				  if(tmp_counter<uc_ntrylimit){
				          /* it shouldn't be connected to
					   * lv[iv1] 
					   */
				          connected = 0;
					  j=1;
					  while(j<=degree[lv[iv1]]){
					          if(NEI[lv[iv1]][j]==jv1){
						          connected = 1;
							  /* exit j loop */
							  j=degree[lv[iv1]]+1;
						  }
						  else{
						          j++;
						  }
					  }
				  }

			  }while(  (tmp_counter<uc_ntrylimit)
				 &&(connected==1)
				);



			  /* if appropriate first vertex of the second
			   * pair has been found, look for second
			   * vertex of second pair 
			   */ 
			  /* next three lines dummy */
			  if(tmp_counter>=uc_ntrylimit){ 
			          jv2 = -1;
			  }
			  else{ /* ie. if(tmp_counter<uc_ntrylimit) */
			          tmp_counter = 0;
				  do{
				          /* pick a vertex differing
					   * from the previous three 
					   */
				          do{
					          jv2=1+(int)floor((double)uc_nvertex*ranksz());
						  tmp_counter++;
					  }while(  (tmp_counter<uc_ntrylimit)
						 &&(  (jv1==jv2)
						    ||(lv[iv1]==jv2)  
						    ||(lv[iv2]==jv2)  
						   )
						);		     

					  /* if found such a vertex: */
					  if(tmp_counter<uc_ntrylimit){
					          /* it should be connected to
						   * jv1, but not to lv[iv2]
						   */
					          connected_to_jv1 = 1;
						  j=1;
						  while(j<=degree[jv1]){
					                  if(NEI[jv1][j]==jv2){
							          connected_to_jv1 = 0;
								  /* exit j loop */
								  j=degree[jv1]+1;
							  }
							  else{
							          j++;
							  }
						  }

						  connected_to_lv_iv2 = 0;
						  j=1;
						  while(j<=degree[lv[iv2]]){
						          if(NEI[lv[iv2]][j]==jv2){
							          connected_to_lv_iv2 = 1;
								  /* exit j loop */
								  j=degree[lv[iv2]]+1;
							  }
							  else{
							          j++;
							  }
						  }	  
					  }

				  }while(  (tmp_counter<uc_ntrylimit)
					 &&(  (connected_to_jv1==0)
					    ||(connected_to_lv_iv2==1)
					   )
					);
			  }



			  /* see, if we've succeded */
			  if(tmp_counter>=uc_ntrylimit){
			          exit_flag=0;
			  }

			  /* if not, try again */
		  }while(exit_flag==0);




		  /* disconnect jv1 and jv2 
		   * connect lv[iv1] and jv1 
		   * connect lv[iv2] and jv2
 		   */
		  

		  /* the valency jv1 -> jv2 is moved to jv1 -> lv[iv1] */
		  j = 1;
		  while(  (j<=c_lnl[jv1]-1)
			&&(NEI[jv1][j]!=jv2)
		       ){
		          j++;
		  }
		  if(j>c_lnl[jv1]-1){
		          fprintf(stderr,"ERROR / jv1.0\n");
			  MY_EXIT;
		  }
		  index_of_jv2_on_jv1s_list = j;
		  NEI[jv1][index_of_jv2_on_jv1s_list] = lv[iv1];

		  /* the free valency iv1 of lv[iv1] is connected to jv1 */
		  NEI[lv[iv1]][c_lnl[lv[iv1]]] = jv1;
		  c_lnl[lv[iv1]]++;


		  /* the valency jv2 -> jv1 is moved to jv2 -> lv[iv2] */
		  j = 1;
		  while(  (j<=c_lnl[jv2]-1)
			&&(NEI[jv2][j]!=jv1)
		       ){
		          j++;
		  }
		  if(j>c_lnl[jv2]-1){
		          fprintf(stderr,"ERROR / jv2.0\n");
			  MY_EXIT;
		  }
		  index_of_jv1_on_jv2s_list = j;
		  NEI[jv2][index_of_jv1_on_jv2s_list] = lv[iv2];

		  /* the free valency iv2 of lv[iv2] is connected to jv2 */
		  NEI[lv[iv2]][c_lnl[lv[iv2]]] = jv2;
		  c_lnl[lv[iv2]]++;



		  /* updating book-keeping of free valencies */
		  tmp_iv1 = iv1;
		  iv1 = (int)MIN(iv1,iv2);
		  iv2 = (int)MAX(tmp_iv1,iv2);
		  for(iv=iv2;iv<=novR-1;iv++){ lv[iv]=lv[iv+1]; }
		  novR--;
		  for(iv=iv1;iv<=novR-1;iv++){ lv[iv]=lv[iv+1]; }
		  novR--;


		  /* number of edges increased by 1 */
		  (*uc_nedge)++;
	          break;          
	  }
	  }
  }



  /* special case:
   * if all remaining valencies are on one vertex:
   * 1) stop the algorithm, it's only one vertex, will not harm
   * 2) change the degree[this vertex] appropriately:
   *    _rarely_, this may cause this special vertex to have
   *    degree[i]=0, ie. it will not be connected to the rest of the
   *    graph 
   */
  if( novR > 0 ){
          i = lv[1];
	  degree[i] -= novR;
  }

  

  free_lvector(c_lnl,1,uc_nvertex);
  free_lvector(lv,1,nov);
}

/*-----------------------------------------*/

void GenerateMatrix_SW( int seed, int model,
			unsigned long sw_nvertex,
			int sw_k,
			double sw_p,
			unsigned long *degree
			){

  /* generating the adjacency matrix of a small-world graph;
   * Nature v393 p440-443
   * originally, every node is connected to the sw_k nearest
   * neighbors, and sw_p is the rewiring probability
   */
  unsigned long sw_nedge,i,*pam,j,tmpul,*nei_ncl,*nei_nch,
    *c_lnl,ie,iv;
  int connected,dist,exit_flag;

  /* 
   * ie: index for edges
   * iv: index for vertices
   * dist: distance
   *
   * (pam[ie],pam[sw_nedge+ie]) are the vertices connected by the (ie)th edge,
   *                   pam = pair member
   * c_lnl: counter for the length of the neighbor list
   *        c_lnl[i]-1 is the number of neighbors of vertex i
   * nei_ncl[i] is the starting(l=low) index of the ith row of NEI[][]
   *     nch is the last(h=high)
   */



  /* 0.
   * init 
   */
  seedksz(seed);
  sw_nedge = sw_nvertex * sw_k/2;
  pam=lvector(1,2*sw_nedge);
  c_lnl=lvector( 1, sw_nvertex );
  nei_ncl=lvector( 1, sw_nvertex );
  nei_nch=lvector( 1, sw_nvertex );




  /* 1.
   * initializing pair list of original (non-rewired) graph
   * [sw_k is an even positive number]
   * initializing degree list also
   */
  ie=1;
  for(dist=1;dist<=sw_k/2;dist++){
          for(iv=1;iv<=sw_nvertex;iv++,ie++){
	          pam[ie]=iv;
		  pam[sw_nedge+ie]=((iv+dist)+sw_nvertex-1)%sw_nvertex+1;
	  }
  }
  for(iv=1;iv<=sw_nvertex;iv++){ degree[iv] = sw_k; }


  

  /* 2.
   * rewiring in edge list
   */
  /* going around the circle clockwise */
  ie=1;
  for(iv=1;iv<=sw_nvertex;iv++){
          /* checking the clockwise (dist)th nearest neighbor of vertex iv */
          for(dist=1;dist<=sw_k/2;dist++,ie++){

	        /* with probability sw_p, rewire */
	        if(ranksz()<=sw_p){
		    /* pick a new point for neighbor
		     * should not be connected to iv, and not be
		     * identical to the present neighbor
		     */
		    exit_flag = 0; /* whether good new point has been found */
		    do{
		        do{
			    tmpul=1+(unsigned long)floor((double)sw_nvertex*ranksz());
			}while((tmpul==iv)||(tmpul==pam[ie+sw_nedge]));
		  
			
			connected = 0;
			/* whether iv and tmpul are already connected */
			/* first, check for the connections of iv */
			j=1;
			while(j<=sw_k/2){
			        if(  (j!=dist) 
				     /* exclude connection presently
				      * to be changed from investigation
				      */
				   &&(pam[iv*sw_k/2+j+sw_nedge]==tmpul)
				  ){
				        connected = 1;
					j=sw_k/2; /* exit j loop */
				}
				j++;
			}
			/* check for the connections of tmpul */
			j=1;
			while(j<=sw_k/2){
			        if(pam[tmpul*sw_k/2+j+sw_nedge]==iv){
				        connected = 1;
					j=sw_k/2; /* exit j loop */
				}
				j++;
			}
			if(connected==0){
			        exit_flag = 1;
			}
		    }while(exit_flag==0);			  


		    /* changing degrees of vertices */
		    /* decrementing the degree of the vertex that was
		     * connected to iv until now
		     */
		    degree[pam[ie+sw_nedge]]--;
		    /* incrementing the degree of the vertex that will
		     * be connected to iv now
		     */
		    degree[tmpul]++;


		    /* change entry belonging to altered end of
		     * edge in edge list 
		     */
		    pam[ie+sw_nedge]=tmpul;
	      }
	  }
  }
  


  /* 3.
   * transforming the adjacency matrix from the previous pam
   * into row list format:
   * the non-zero entries (all of them being 1) of row i 
   * will be listed in the array NEI[i]
   */
  
  /* 3.1 init */
  for(iv=1;iv<=sw_nvertex;iv++){
          nei_ncl[iv]=1;
	  nei_nch[iv] = degree[iv];
  }
  NEI=lmatrix_vrl(1,sw_nvertex,nei_ncl,nei_nch);

  /* 3.2 creating neighbor lists */
  for(iv=1;iv<=sw_nvertex;iv++){
	  c_lnl[iv]=1;
  }
  for(ie=1;ie<=sw_nedge;ie++){
          i = pam[ie];
          j = pam[sw_nedge+ie];
          NEI[i][c_lnl[i]] = j;
	  c_lnl[i]++;
          NEI[j][c_lnl[j]] = i;
	  c_lnl[j]++;
  }


  free_lvector(pam,1,2*sw_nedge);
  free_lvector( c_lnl, 1, sw_nvertex );
  free_lvector( nei_ncl, 1, sw_nvertex );
  free_lvector( nei_nch, 1, sw_nvertex );
}

/**********************************************/

void GetEigVals( unsigned long n, int fold_sw, 
		 double spec_shift, int spec_stretch,
		 double folding_point, unsigned long *degree,
		 int *return_n_of_iter,
		 double *return_lambda, double *return_beta, 
		 double *return_error, double *return_sum_omega_abs,
		 double *ipr
		 ){

  /**********************************************************
   * switch(fold_sw){
   * case 0:{   PLAIN
   *      the top Lanz_Nd eigenvalues of the matrix stored in
   *      NEI and degree
   * }
   * case 4:{     PLAIN
   *      the lowest Lanz_Nd eigvals of the matrix stored in 
   *      NEI and degree
   *      i.e., the top eigvals of the matrix -A
   * }
   * case 1:{   FOLD/SUBSTRACT/SHIFT
   *      look for the Lanz_Nd largest eigvals of  
   *              spec_shift*I - (A-folding_point*I)^2
   *      and with the eigenvectors obtained compute the Lanz_Nd
   *      eigenvalues of the matrix that are closest to the folding point 
   * }
   * case 2:{   FOLD
   *      look for the largest eigval of   (A-folding_point*I)^2
   *      which gives the lowest eigenvalue of    - (A-folding_point*I)^2
   * }
   * case 3:{   FOLD & STRETCH
   *      let B = shift - (A-folding_point*I)^2
   *      look for the largest Lanz_Nd eigvals of B^{spec_stretch}
   *      then, with the eigenvectors, compute the Lanz_Nd eigvals of
   *      the original A matrix that are closest to folding_point
   * }
   * case 10:{   FOLD/SUBSTRACT/SHIFT
   *      similar to 1, difference: folding point differs from the
   *      point where the eigvals are looked for
   * }
   * case 30:{   FOLD & STRETCH
   *      similar to 3, difference: folding point differs from the
   *      point where the eigvals are looked for
   * }
   * }
   *--------------------------------------*
   * Using the THICK-RESTART LANCZOS METHOD
   * (http://www.nersc.gov/~kewu/)
   *
   * K Wu / A Canning / H D Simon: LBNL-42917
   * K Wu / H D Simon: LBNL-41412
   **********************************************************/


  FILE *fp[8];
  char fn[8][MY_STRLEN];
  unsigned long i,j,l,*indx_d_tqli,*new_indx,ii,
    n_of_iter,l1,l2,tmpul,tmp_ul0,tmp_ul1;
  int exit_flag,first_start_flag,lanczos_k,random_q_flag,
    n_of_converged,tmpk,*debug,number_of_eigpairs_to_save,
    foldswitch1030_new_k;
  double **q,qsqr,one_over_sqrt_qsqr,*beta,*alpha,beta_q,eta,qq1,qq2,
    one_over_beta,**a_tqli,*e_tqli,*d_tqli,*erp,*tmp_nvec,
    *szendvics,tmp_sum_rsqr,beta_m,*tmp_mvec,*ritz_vec_sqr,
    **omega,sum_omega_abs,largest_omega_abs,tmpd,tmpd2,
    tmp_alpha;
  float *tmp_fnvec;
  
  /* erp: relative error of Ritz pair
   * n: number of vertices
   * n_of_iter: number of iterations so far
   * omega[i][j] = q_i q_j  (a measure of error)
   * ipr: inverse participation ratio (of eigenvectors)
   */
  


  /* check parameters */
  if(  ((fold_sw==3)||(fold_sw==30))
     &&((spec_stretch<=0)||(spec_stretch%2==0))
    ){
          fprintf(stderr,"PARAMETER ERROR in ''GetEigVals'':\n");
	  fprintf(stderr,"''spec_stretch'' has to be an odd positive number\n");
	  MY_EXIT;
  }


  /* init debug files */
  debug=ivector(0,7);
  for(i=0;i<=7;i++){
          debug[i] = ( Lanz_DebugSwitch&((int)(pow(2.0,i))) )  >> i;
	  if(debug[i]==1){
	          sprintf(fn[i],"%s_%ld_seed%d_size%ld_",
			  Lanz_DebugName,i,RndSeed,n);
	          fp[i] = fopen(fn[i],"w");
	  }
  }


  /* memory allocation */
  q=dmatrix(1,Lanz_M+1,1,n);
  beta=dvector(1,Lanz_M);
  alpha=dvector(1,Lanz_M);
  d_tqli=dvector(1,Lanz_M);
  e_tqli=dvector(1,Lanz_M);
  a_tqli=dmatrix(1,Lanz_M,1,Lanz_M);
  indx_d_tqli=lvector(1,Lanz_M);
  erp=dvector(1,Lanz_M); 
  szendvics=dvector(1,Lanz_M); 
  tmp_nvec=dvector(1,n);
  new_indx=lvector(1,Lanz_M);
  tmp_mvec=dvector(1,Lanz_M);
  ritz_vec_sqr=dvector(1,Lanz_M);
  omega=dmatrix(1,Lanz_M+1,1,Lanz_M+1);



  /* init internal variables */
  exit_flag = 0;
  first_start_flag = 1;
  lanczos_k = 0;
  n_of_converged = 0;
  n_of_iter = 0;



  /* ***********   LOOP:  THICK-RESTART LANCZOS ITERATION   *********** */
  while(exit_flag==0){


          /* 0. 
	   * INITIALIZATION (at first start)
	   * or nothing (at restarts)
	   */

          /* 0.1 if this is the first start, init q_1 */
          if(first_start_flag==1){
	          first_start_flag = 0;

		  /* starting vector */
		  /* components distributed uniformly between
		   * -0.5 and 0.5
		   */
		  for(j=1;j<=n;j++){
		          q[1][j] = 0.5-ranksz();
		  }


		  /* normalizing the starting vector */
		  qsqr = 0.0;
		  for(j=1;j<=n;j++){
		          qsqr += MY_SQR(q[1][j]);
		  }
		  one_over_sqrt_qsqr=1.0/sqrt(qsqr);
		  for(j=1;j<=n;j++){
		          q[1][j] *= one_over_sqrt_qsqr;
		  }

		  /* initialize beta[1], even though it's not needed */
		  /*		  beta[1] = 0.0;*/
	  }
	  /* 0.2 at restarts, no init */
	  else{}


 


	  /* 1.
	   * ITERATION
	   */
	  for(i=lanczos_k+1;i<=Lanz_M;i++){


	          /* 1.1 
		   * q_{i+1}
		   */
	          switch(fold_sw){
		  case 0: default:{ 
		          /* searching for the eigvals of the original matrix 
			   * q_{i+1} = A q_{i} 
			   */
		          for(j=1;j<=n;j++){
			          q[i+1][j] = 0.0;
				  l=1;
				  while(l<=degree[j]){
				          q[i+1][j] += q[i][NEI[j][l]];
					  l++;
				  }
			  }
			  break;
		  }
		  case 4:{ 
		          /* searching for lowest eigvals of the
			   * original matrix
			   * i.e., for the top the eigvals of the
			   * matrix -A
			   *
			   * q_{i+1} = -A q_{i} 
			   */
		          for(j=1;j<=n;j++){
			          q[i+1][j] = 0.0;
				  l=1;
				  while(l<=degree[j]){
				          q[i+1][j] -= q[i][NEI[j][l]];
					  l++;
				  }
			  }
			  break;
		  }
		  case 1: case 10:{
			  /* spec_shift*I  -  ( A - I*folding_point )^2
			   *  = I*(spec_shift-folding_point^2) 
			   *    + 2*A*folding_point - A^2
			   */

		          /* computing   A q_i:  storing in tmp_nvec */
		          for(j=1;j<=n;j++){
				  tmp_nvec[j] = 0.0;
				  l=1;
				  while(l<=degree[j]){
				          tmp_nvec[j] += q[i][NEI[j][l]];
					  l++;
				  }
			  }
			  /* computing -A^2 q_i: storing in q_{i+1}*/
		          for(j=1;j<=n;j++){
				  l=1;
				  while(l<=degree[j]){
				       q[i+1][j] = -1.0*tmp_nvec[NEI[j][l]];
				       l++;
				  }
			  }
			  /* adding 2*A*folding_point */
			  tmpd = 2.0*folding_point;
			  for(j=1;j<=n;j++){
			          q[i+1][j] += tmpd*tmp_nvec[j];
			  }
			  /* adding   
			   * (spec_shift-folding_point^2) * q_i 
			   */
			  tmpd = spec_shift - MY_SQR(folding_point);
			  for(j=1;j<=n;j++){
			          q[i+1][j] += tmpd*q[i][j];
			  }
			  break;
		  }
		  case 2:{
		          /* searching for the largest eigval of 
			   * ( A - I*folding_point )^2
			   * = I * folding_point^2 - 2*A*folding_point + A^2
			   */

		          /* computing   A q_i:  storing in tmp_nvec */
		          for(j=1;j<=n;j++){
				  tmp_nvec[j] = 0.0;
				  l=1;
				  while(l<=degree[j]){
				          tmp_nvec[j] += q[i][NEI[j][l]];
					  l++;
				  }
			  }
			  /* computing  A^2 q_i:  storing in q_{i+1} */
		          for(j=1;j<=n;j++){
			          q[i+1][j] = 0.0;
				  l=1;
				  while(l<=degree[j]){
				          q[i+1][j] += tmp_nvec[NEI[j][l]];
					  l++;
				  }
			  }
			  /* adding -2*A*folding_point */
			  for(j=1;j<=n;j++){
			          q[i+1][j] -= 2.0*folding_point*tmp_nvec[j];
			  }
			  /* adding folding_point^2 * q_i */
			  for(j=1;j<=n;j++){
			          q[i+1][j] += MY_SQR(folding_point)*q[i][j];
			  }
			  break;
		  }
		  case 3: case 30:{
		          /* searching for largest Lanz_Nd eigvals of the
			   * folded & enhanced matrix, which is B^{spec_stretch}
			   *
			   * B = spec_shift I - ( A - folding_point I )^2 
			   *   = ( x_0 I + x_1 A ) * ( y_0 I + y_1 A )
			   *
			   * and B^{spec_stretch}
			   *     = sum_{j=0...2*spec_stretch} A^j * a_coeff[j],
			   * where a_coeff[j] = sum_{l=l1...l2} 
		    * x1^l x0^{spec_stretch-l} (spec_stretch\atop l)
		    * y1^{j-l} y0^{spec_stretch-(j-l)} (spec_stretch\atop j-l)
		           * and l1=max(0,j-spec_stretch),
			   *     l2=min(j,spec_stretch)
		           */

		          double x0,x1,y0,y1,**binom,*tmp_nvec_2,*a_coeff;
			  int j,k,l,l1,l2;
			  /* binom[j][k] = j!/(k!*(j-k)!)  */
			  binom=dmatrix(0,spec_stretch,0,spec_stretch);
			  tmp_nvec_2=dvector(1,n);
			  a_coeff=dvector(0,2*spec_stretch);

			  
			  /* init */
                          /* constants */ 
			  x0 = sqrt(spec_shift) - folding_point;
			  x1 = 1.0;
			  y0 = sqrt(spec_shift) + folding_point;
			  y1 = -1.0;
			  /* binomial coefficients */
			  for(j=0;j<=spec_stretch;j++){
			          for(k=0;k<=j;k++){
				          /* init */
				          binom[j][k] = 1.0;
					  /* enumerator */
					  l=1;
					  while(l<=j){
					          binom[j][k]*=(double)l;
						  l++;
					  }
					  /* denominator */
					  l=1; 
					  while(l<=k){
					          binom[j][k]/=(double)l;
						  l++;
					  }
					  l=1;
					  while(l<=j-k){
					          binom[j][k]/=(double)l;
						  l++;
					  }
				  }
			  }
			  /* B^{spec_stretch} is a polinomial of A, 
			   * here come the coefficients of the
			   * different powers of A in this polinomial
			   */
			  for(j=0;j<=2*spec_stretch;j++){
			          a_coeff[j] = 0.0;
				  l1=MY_MAX(0,j-spec_stretch);
				  l2=MY_MIN(j,spec_stretch);
				  l=l1;
				  while(l<=l2){
				          a_coeff[j] += 
					    pow(x1,(double)l)
					    *pow(x0,(double)(spec_stretch-l))
					    *binom[spec_stretch][l]
					    *pow(y1,(double)(j-l))
					    *pow(y0,(double)(spec_stretch-j+l))
					    *binom[spec_stretch][j-l];
					  l++;
				  }
			  }

			  
			  /* computing the polynomial */
			  /* init */
			  for(j=1;j<=n;j++){ 
			          tmp_nvec[j] = q[i][j]; 
				  q[i+1][j] = 0.0;
			  }
			  /* loop */
			  for(k=0;k<=2*spec_stretch;k++){
			          /* copying   A^k q_i   into   tmp_nvec_2 */
			          for(j=1;j<=n;j++){ 
				          tmp_nvec_2[j] = tmp_nvec[j]; 
				  }
				  /* adding  a_coeff[k]*A^k q_i  to  q_{i+1} */
				  for(j=1;j<=n;j++){
				          q[i+1][j] += a_coeff[k]*tmp_nvec_2[j];
				  }
				  /* computing  A^{k+1} q_i
				   * and storing it in  tmp_nvec
				   * [not needed in last loop]
				   */
				  if(k<2*spec_stretch){
				          for(j=1;j<=n;j++){
					          l=1;
					          tmp_nvec[j] = 0.0;
						  while(l<=degree[j]){
						        tmp_nvec[j] += 
							  tmp_nvec_2[NEI[j][l]];
							l++;
						  }
					  }
				  }
			  }
			  

			  /* cleaning up */
			  free_dvector(tmp_nvec_2,1,n);
			  free_dmatrix(binom,0,spec_stretch,0,spec_stretch);
			  free_dvector(a_coeff,0,2*spec_stretch);
		          break;
		  }
		  }



		  /* 1.2
		   * alpha[i] = q^T_{i} q_{i+1}
		   */
		  alpha[i] = 0.0;
		  for(j=1;j<=n;j++){
		          alpha[i] += q[i][j]*q[i+1][j];
		  }
		  


		  /* 1.3 orthogonalization */

		  if(i>lanczos_k+1){
		          for(j=1;j<=n;j++){
			          q[i+1][j] -=   alpha[i]*q[i][j] 
				               + beta[i-1]*q[i-1][j];
			  }
		  }
		  else{
		          for(j=1;j<=n;j++){
			            beta_q = 0.0;
				    l=1;
				    while(l<=lanczos_k){
				              beta_q += beta[l]*q[l][j];
					      l++;
				    }
				    q[i+1][j] -= alpha[i]*q[i][j] + beta_q; 
			  } 
		  }		  
		  


/* checking separately how orthogonal q_{i+1} is to the previous q
 * vectors 
 */
if(debug[5]==1){
  fprintf(fp[5],"#B\tn_iter\ti\tj\tq_j q_{i+1}\n");
  l=1;
  while(l<i){
    tmpd=0.0;
    for(j=1;j<=n;j++){
      tmpd+=q[l][j]*q[i+1][j];
    }
    fprintf(fp[5],"\t%ld\t%ld\t%ld\t%12.10E\n",n_of_iter,i,l,tmpd);
    l++;
  }
  fprintf(fp[5],"\n");
  /*fflush(fp[5]);*/
}



		  /* 1.4 re-orthogonalization */
                  /* 1.4.a: eta */
                  if(i>lanczos_k+1){ 
		          eta = MY_SQR(alpha[i]) + MY_SQR(beta[i-1]);
		  }
		  else{
		          eta = MY_SQR(alpha[i]);
			  l = 1;
			  while(l<=lanczos_k){
			          eta += MY_SQR(beta[l]);
				  l++;
			  }
		  }
		  /* 1.4.b:  q^T_{i+1} q_{i+1} = qsqr */
		  qsqr = 0.0;
		  for(j=1;j<=n;j++){ 
		          qsqr += MY_SQR(q[i+1][j]);
		  }	      
			  
			  

		  /* 1.4.c: the re-orthogonalization step */
		  /* also: telling the normalization step, whether q[i+1]
		   * was replaced by a random vector in this step 
		   */ 
		  if(qsqr>=eta){

		          /* first, update alpha[i]:
			   * alpha[i] += q^T_i q_{i+1}
			   */
		          for(j=1;j<=n;j++){
			          alpha[i] += q[i][j]*q[i+1][j];
			  }
			  /* then, perform the LOCAL re-orthogonalization */

			  if(i>1){
			          qq1 = qq2 = 0.0;
				  for(j=1;j<=n;j++){
				          qq1 += q[i][j]*q[i+1][j];
					  qq2 += q[i-1][j]*q[i+1][j];
				  }
				  for(j=1;j<=n;j++){
				          q[i+1][j] -= q[i][j]*qq1+q[i-1][j]*qq2;
				  }
			  }
			  else{
			          qq1 = 0.0;
				  for(j=1;j<=n;j++){
				          qq1 += q[i][j]*q[i+1][j];
					  /* there's no q_{i-1}, if i=1 */
				  }
				  for(j=1;j<=n;j++){
				          q[i+1][j] -= q[i][j]*qq1;
				  }		      
			  }
			  
			  /* no random q_{i+1} was used */
			  random_q_flag = 0;
		  }
		  else if(qsqr>=eta*MY_SQR(Lanz_MEps)){

		          /* first, update alpha[i]:
			   * alpha[i] += q^T_i q_{i+1}
			   */
		          for(j=1;j<=n;j++){
		                  alpha[i] += q[i][j]*q[i+1][j];
			  }
			  /* then, perform the GLOBAL re-orthogonalization */
			  l = 1;
			  while(l<=i){
			          qq1 = 0.0;
				  for(j=1;j<=n;j++){
				          qq1 += q[l][j]*q[i+1][j];
				  }
				  for(j=1;j<=n;j++){
				          q[i+1][j] -= q[l][j]*qq1;
				  }
				  l++;
			  }
			  
			  /* no random q_{i+1} was used */
			  random_q_flag = 0;
		  }
		  else{

		          /* replacing q[i+1] by a random vector that is
			   * orthogonal to q[1]...q[i]
			   */

		          /* random vector:
			   * components distributed uniformly between
			   * -0.5 and 0.5
			   */
		          for(j=1;j<=n;j++){
			          q[i+1][j] = 0.5-ranksz();
			  }

			  /* orthogonalizing */
			  l = 1;
			  while(l<=i){
			          qq1 = 0.0;
				  for(j=1;j<=n;j++){
				          qq1 += q[l][j]*q[i+1][j];
				  }
				  for(j=1;j<=n;j++){
				          q[i+1][j] -= q[l][j]*qq1;
				  }
				  l++;
			  }
			  
			  /* random q_{i+1} was used */
			  random_q_flag = 1; 
		  }
		  



		  /* 1.5
		   * normalization
		   */
		  qsqr = 0.0;
		  for(j=1;j<=n;j++){
		          qsqr += MY_SQR(q[i+1][j]);
		  }
		  beta[i] = sqrt(qsqr);
		  one_over_beta = 1.0/beta[i];
		  for(j=1;j<=n;j++){
		          q[i+1][j] *= one_over_beta;
		  }
		  if(random_q_flag==1){
			          beta[i] = 0.0;
		  }
          }


/* testing whether Eq.1 is true after the iteration loop */
if(debug[4]==1){
  fprintf(fp[4],"\n#ITER_END\n");
  fprintf(fp[4],"#n_iter\tm\ti\tsqr(lhs.Eq.1-rhs.Eq.1)\n");
  for(ii=1;ii<=Lanz_M;ii++){
    tmp_sum_rsqr=0.0;
    for(j=1;j<=n;j++){
      tmp_nvec[j]=0.0;
      l=1;
      while(l<=degree[j]){
	tmp_nvec[j]+=q[ii][NEI[j][l]];
	l++;
      }
      tmp_nvec[j]-=alpha[ii]*q[ii][j];
      if(ii>1){tmp_nvec[j]-=beta[ii-1]*q[ii-1][j];}
      if(ii<Lanz_M){tmp_nvec[j]-=beta[ii]*q[ii+1][j];}
      if(ii==Lanz_M){tmp_nvec[j]-=beta[ii]*q[Lanz_M+1][j];}
      tmp_sum_rsqr+=MY_SQR(tmp_nvec[j]);
    }
    fprintf(fp[4],"%ld\t%ld\t%ld\t%f\n",n_of_iter,Lanz_M,ii,tmp_sum_rsqr);
  }
  /*fflush(fp[4]);*/
}





	  /* 2. 
	   * EVALUATION (the restarting scheme)
	   * computing the Ritz values and Ritz vectors
	   * deciding which Ritz values/vectors should be kept
	   * deciding whether to stop
	   */
	  


	  /* 2.1 eigvals and eigvects of T_m */ 
          if(n_of_iter==0){ /* if the matrix is tridiagonal */

                  /* init a_tqli */
                  for(i=1;i<=Lanz_M;i++){
		          a_tqli[i][i] = 1.0;
			  j = i+1;
			  while(j<=Lanz_M){
			          a_tqli[i][j] = a_tqli[j][i] = 0.0;
				  j++;
			  }
		  }
		  
		  /* init d_tqli and e_tqli */
		  for(i=1;i<=Lanz_M;i++){
		          d_tqli[i] = alpha[i];
		  }
		  for(i=2;i<=Lanz_M;i++){
		          e_tqli[i] = beta[i-1];
		  }
			  

		  /* get eigvals and eigvects */
		  tqli_il_d_ul(d_tqli,e_tqli,Lanz_M,a_tqli,tqli_ItMax);
	  }
	  else{ /* T_m is not tridiagonal */

                  /* init a_tqli */
                  for(i=1;i<=Lanz_M;i++){
		          j = i+1;
			  while(j<=Lanz_M){
			          a_tqli[i][j] = a_tqli[j][i] = 0.0;
				  j++;
			  }
		  }
		  for(i=1;i<=Lanz_M;i++){
		          a_tqli[i][i] = alpha[i];
		  }
		  for(i=1;i<=lanczos_k;i++){
		          a_tqli[lanczos_k+1][i] =
			    a_tqli[i][lanczos_k+1] = beta[i];
		  }
		  for(i=lanczos_k+2;i<=Lanz_M;i++){
		          a_tqli[i-1][i] = a_tqli[i][i-1] = beta[i-1];
		  }
			  
		  /* get eigvals and eigvects */		  
		  tred2_d_ul(a_tqli,Lanz_M,d_tqli,e_tqli);
		  tqli_il_d_ul(d_tqli,e_tqli,Lanz_M,a_tqli,tqli_ItMax);
	  }




	  /* 2.2
	   * computing the Ritz vectors (stored in the q vectors)
	   * normalizing the Ritz vectors, if needed
	   */
	  for(i=1;i<=Lanz_M;i++){
	          ritz_vec_sqr[i] = 0.0;
	  }
	  for(j=1;j<=n;j++){
	          /* computing the jth component of every Ritz vector */
	          for(i=1;i<=Lanz_M;i++){
		          tmp_mvec[i] = 0.0;
			  for(l=1;l<=Lanz_M;l++){
			          tmp_mvec[i]+=q[l][j]*a_tqli[l][i];
			  }
		  }
		  /* copying the result into the jth row of the Q
		   * matrix 
		   */
		  for(i=1;i<=Lanz_M;i++){
		          q[i][j] = tmp_mvec[i];
			  ritz_vec_sqr[i] += MY_SQR(tmp_mvec[i]);
		  }
	  }
	  /* normalizing every Ritz vector to length 1, if needed */
	  for(i=1;i<=Lanz_M;i++){
		  if(ritz_vec_sqr[i]!=1.0){
		          one_over_sqrt_qsqr = 1.0/sqrt(ritz_vec_sqr[i]);
		          for(j=1;j<=n;j++){
			          q[i][j] *= one_over_sqrt_qsqr;
			  }
		  }
	  }




	  /* 2.3
	   * fixing new order of indices
	   */
	  switch(fold_sw){
	  case 0: case 1: case 2: case 3: case 4:{

	          /* either no folding or eigvals are looked for 
		   * around the folding point;
		   * with other words: either the top eigvals of the
		   * original spectrum, or the top eigvals of the -A
		   * matrix, or the top eigvals of the folded
		   * spectrum are to be kept
		   */

	          /* the lanczos_k Ritz pairs with the largest Ritz values
		   * shall be kept => the new_indx array is the ascending
		   * order of the eigvals of T_m
		   * (the updated value of lanczos_k is not known yet)
		   */

	          /* index-sorting the eigvals into ___ascending___ order */ 
	          indexx_ov_d_ul(Lanz_M,d_tqli,indx_d_tqli);
		  /* we're looking for the largest eigvals, 
		   * so we'll need the descending order 
		   */
		  for(i=1;i<=Lanz_M;i++){
		          new_indx[i] = indx_d_tqli[Lanz_M+1-i];
		  }

		  break;
	  }
	  case 10: case 30:{

	          double c_lo,c_hi;
		  int *eigin;
		  eigin=ivector(1,Lanz_M);

	          /* folding,
		   * eigvals are taken from a point different from the
		   * folding point;
		   * with other words: eigvals from the interval 
		   *     folding_point - Lanz_Phi - Lanz_IW/2, 
		   *     folding_point - Lanz_Phi + Lanz_IW/2
		   * are to be kept.
		   *
		   * The new_indx array is fixed such that those
		   * eigenpairs of the original matrix that fall into
		   * this interval will be put to the top of
		   * the list and those outside to the bottom.
		   */

		  fprintf(stderr,"Sorry, implementation of Lanz_Fold=10\n");
		  fprintf(stderr,"in GetEigVals/2.2 not finished yet\n");
		  MY_EXIT;


		  c_lo = folding_point - Lanz_Phi - 0.5*Lanz_IW;
		  c_hi = folding_point - Lanz_Phi + 0.5*Lanz_IW;

		  for(i=1;i<=Lanz_M;i++){
		          
		          /* testing if this eigval falls into the interval
			   * [c_lo, c_hi]
			   */
		          for(j=1;j<=n;j++){
			          tmp_nvec[j]=0.0;
				  l=1;
				  while(l<=degree[j]){
				          tmp_nvec[j]+=q[i][NEI[j][l]];
					  l++;
				  }
			  }
		          tmpd=0.0;
			  tmpul=0;
			  for(j=1;j<=n;j++){
			          if(tmp_nvec[j]!=0.0){
				          tmpul++;
				          tmpd+=q[i][j]/tmp_nvec[j];
				  }
			  }
			  tmpd/=(double)tmpul;


			  /* setting flag indicating whether eigval is
			   * inside 
			   */
			  if((tmpd>=c_lo)&&(tmpd<=c_hi)){
			          eigin[i]=1;
			  }
			  else{
			          eigin[i]=0;
			  }
		  }



		  /* # of eigvals to be kept */
		  for(i=1,foldswitch1030_new_k=0;i<=Lanz_M;i++){
		          if(eigin[i]==1){
			          foldswitch1030_new_k++;  
			  }
		  }


		  /* putting indices into the new_indx array */
		  for(i=1;i<=Lanz_M;i++){
		          tmp_ul0=foldswitch1030_new_k+1;
		          tmp_ul1=1;

		          switch(eigin[i]){
			  case 0:{
			          new_indx[tmp_ul1]=i;
				  tmp_ul1++;
				  break;
			  }
			  case 1:{
			          new_indx[tmp_ul0]=i;
				  tmp_ul0++;
				  break;
			  }
			  }

		  }


		  free_ivector(eigin,1,Lanz_M);
		  break;
	  }
	  }
  
		  


	  /* 2.3 */
	  /* rearranging the Ritz vectors */
		  for(j=1;j<=n;j++){
		          for(i=1;i<=Lanz_M;i++){
			          tmp_mvec[i] = q[new_indx[i]][j];
			  }
			  for(i=1;i<=Lanz_M;i++){
			          q[i][j] = tmp_mvec[i];
			  }
		  }






          /* 2.4 errors of Ritz pairs 
	   *
	   * !!! computed only if the largest eigval of the original
	   *     matrix is being looked for
	   *
	   * erp_i = | x^T_i A x_i | / | x^T_i x_i | - lambda_i
	   * The Ritz vectors have all been normalized to length 1 in
	   * the previous step 
	   */
	  switch(fold_sw){
	  case 0: default:{
	          for(i=1;i<=Lanz_M;i++){
		          szendvics[i] = 0.0;
			  for(j=1;j<=n;j++){
			          tmp_nvec[j] = 0.0;
				  l=1;
				  while(l<=degree[j]){
				          tmp_nvec[j] += q[i][NEI[j][l]];
					  l++;
				  }
				  szendvics[i] += q[i][j]*tmp_nvec[j];
			  }
			  erp[i] = szendvics[i] - d_tqli[new_indx[i]];
		  }
		  break;
	  }
	  case 1: case 2: case 3: case 10: case 30: case 4:{
	          break;
	  }
	  }


/* checking whether q_1 and q_2 are orthogonal */
/*
 *tmpd=0.0;
 *for(j=1;j<=n;j++){
 *tmpd+=q[1][j]*q[2][j];
 *}
 *fprintf(stderr,"%ld\t%8.4E\n",n_of_iter,tmpd);fflush(stderr);
 */


/* saving the alpha, beta and erp values */
if(debug[2]==1){
  switch(fold_sw){
  case 0: default:{/*when searching for the eigvals of the original matrix*/
    fprintf(fp[2],"\n\n#i/nit\talpha_old\tbeta_old\teigval_new\terp\n");
    for(i=1;i<=Lanz_M;i++){
      fprintf(fp[2],"%ld %ld\t%8.4E\t%8.4E\t%8.4E\t%8.4E\n",
	      i,n_of_iter,alpha[i],beta[i],d_tqli[new_indx[i]],erp[i]);
    }
    /*fflush(fp[2]);*/
    break;
  }
  case 1: case 2: case 3: case 10: case 30: case 4:{
    fprintf(fp[2],"---\n");
    break;
  }
  }
}


/* saving only the beta's (the residuals) and the new alpha's */
if(debug[7]==1){
  switch(fold_sw){
  case 0: default:{/*when searching for the eigvals of the original matrix*/
    fprintf(fp[7],"\n\n# nit\ti\t\talpha\terp\n");
    for(i=1;i<=Lanz_M;i++){
      fprintf(fp[7],"%ld\t%ld\t%8.4E\t%8.4E\n",
	      n_of_iter,i,d_tqli[new_indx[i]],erp[i]);
    }
    break;
  }
  case 1: case 2: case 3: case 4: case 10: case 30:{
    fprintf(fp[7],"---\n");
    break;
  }
  }
}




	  /* 2.6
	   * updating lanczos_k
	   */

	  /* 
	   * here comes the "magic" formula for updating lanczos_k, 
	   * which is the variable "k" in the notation of Wu et.al.
	   * ["magic" = empirical]
	   * here we're taking the top eigenvalues, so we need
	   * (with the notation of Wu et.al.) the top   m+1 - k_r   values 
	   * (k_l = 0, we don't keep the low eigvals)
	   * later, the eigvals will be put into descending order,
	   * and the new k is the number of Ritz pairs kept
	   */

           /* VERSION 1 */
           /*
	    *	  tmpk = lanczos_k;
	    *	  lanczos_k = Lanz_M+1 - ( 0 + (int)floor(MIN( Lanz_M-Lanz_Nd,
	    *						       0.4*(Lanz_M-n_of_converged) 
	    *						       )
	    *						  )
	    *				   );
	    */
           /* VERSION 2 */
	  tmpk = lanczos_k;
	  switch(fold_sw){
	  case 0: case 1: case 2: case 3: case 4:{
	          if(Lanz_M<=10){ 
		          lanczos_k=(int)ceil(MAX(0.49*Lanz_M,0.99*Lanz_Nd)); 
		  }
		  else{
		          lanczos_k = Lanz_M+1 - ( 0 + (int)floor(MIN( Lanz_M-Lanz_Nd,
								       0.4*(Lanz_M-n_of_converged) 
								       )
								  )
						   );
		  }
		  break;
	  }
	  case 10: case 30:{
	          lanczos_k = foldswitch1030_new_k;
		  break;
	  }
	  }


	  /* 2.7
	   * leaving the chosen (lanczos_k) Ritz vectors in the
	   * q vectors, and putting the (m+1)st q vector into the
	   * (lanczos_k+1)st q vector
	   * updating the alpha and beta arrays, too
	   */

	  /* 2.7.a 
	   * new q vectors
	   */
	  for(j=1;j<=n;j++){
	          q[lanczos_k+1][j] = q[Lanz_M+1][j];
	  }


	  /* 2.7.b
	   * updating alpha and beta
	   */
	  i = 1;
	  beta_m = beta[Lanz_M];
	  while( i <= lanczos_k ){
		  alpha[i] = d_tqli[new_indx[i]];
                  beta[i] = beta_m * a_tqli[Lanz_M][new_indx[i]];
		  i++;
	  }




	  
	  /* 2.8 
	   * testing whether the lanczos_k+1 q vectors are orthogonal 
	   */

	  /* 2.8.a
	   * omega[l1][l2] for every l1,l2=1..i+1
	   */
	  l1=1;
	  while(l1<=lanczos_k+1){
	          l2=l1;
		  while(l2<=lanczos_k+1){
		          omega[l1][l2]=0.0;
			  for(j=1;j<=n;j++){
			          omega[l1][l2] += q[l1][j]*q[l2][j];
			  }
			  omega[l2][l1] = omega[l1][l2];
			  l2++;
		  }
		  l1++;
	  }

	  /* 2.8.b
	   * computing the |omega| sum, and the largest |omega| 
	   */
	  sum_omega_abs = 0.0;
	  largest_omega_abs = -1.0;
	  for(ii=1;ii<=lanczos_k+1;ii++){
	          j=ii+1;
		  while(j<=lanczos_k+1){
		          sum_omega_abs += MY_ABS(omega[ii][j]);
			  if(MY_ABS(omega[ii][j])>largest_omega_abs){
			          largest_omega_abs = MY_ABS(omega[ii][j]);
			  }
			  j++;
		  }
	  }

if(debug[1]==1){
  fprintf(fp[1],"\n# n_iter\ti\tsum |omega|\n");
  fprintf(fp[1],"%ld\t\t%ld\t%8.4E\n",n_of_iter,i,sum_omega_abs);
  /*fflush(fp[1]);*/
}


          /* 2.9.c
	   * if needed:
	   * orthogonalizing the kth vector against the 1...(k-1)st
	   * and orthogonalizing the (k+1)st vector against the 1...kth
	   */

	  if(sum_omega_abs>=Lanz_TeGS){

	          for(i=1;i<=lanczos_k-1;i++){
		          qq1=0.0;
			  for(j=1;j<=n;j++){
			          qq1+=q[i][j]*q[lanczos_k][j];
			  }
			  for(j=1;j<=n;j++){
			          q[lanczos_k][j]-=qq1*q[i][j];
			  }
		  }

	          for(i=1;i<=lanczos_k;i++){
		          qq1=0.0;
			  for(j=1;j<=n;j++){
			          qq1+=q[i][j]*q[lanczos_k+1][j];
			  }
			  for(j=1;j<=n;j++){
			          q[lanczos_k+1][j]-=qq1*q[i][j];
			  }
		  }
	  }




	  /* CHECKING WHETHER FINISHED */
          n_of_iter++;
	  if(n_of_iter>=Lanz_NIterMin){
	          tmpd = 0.1*Lanz_REps;
		  for(i=1;i<=Lanz_Nd;i++){
		          if(MY_ABS(beta[i])>tmpd){
			          tmpd = MY_ABS(beta[i]);
			  }
		  }
		  if( tmpd < Lanz_REps ){
	                  exit_flag = 1;
		  }
	  }


  
  /* testing whether Eq.1 is true at restart */
if(debug[4]==1){
  fprintf(fp[4],"\n#RESTART\n");
  fprintf(fp[4],"\n#n_iter\tk\ti\tsqr(lhs.Eq.1-rhs.Eq.1)\n");
  for(ii=1;ii<=lanczos_k;ii++){
    tmp_sum_rsqr=0.0;
    for(j=1;j<=n;j++){
      tmp_nvec[j]=0.0;
      l=1;
      while(l<=degree[j]){
	tmp_nvec[j]+=q[ii][NEI[j][l]];
	l++;
      }
      tmp_sum_rsqr+=MY_SQR(tmp_nvec[j]-alpha[ii]*q[ii][j]-beta[ii]*q[lanczos_k+1][j]);
    }
    fprintf(fp[4],"%ld\t%ld\t%ld\t%f\n",n_of_iter,lanczos_k,ii,tmp_sum_rsqr);
  }
  /*fflush(fp[4]);*/
}



         /****************** END OF MAIN ITERATION LOOP *************
	 * starting with      ''while(exit_flag==0)''
	 ************************************************************/
  }

  
  





  /* 3. SAVING RETURN VALUES AND CLEANING MEMORY */

  /* 3.0 */
  /* saving inverse participation ratios */
  if(MainSwitch==20){
          for(i=1;i<=Lanz_M;i++){
	          ipr[i]=0.0;
		  for(j=1;j<=n;j++){
		          ipr[i]+=SQR(SQR(q[i][j]));
		  }
	  }
  }



  /* 3.1 */
  switch(fold_sw){
  case 0: case 2: case 4: default:{ 
          /* when looking for the largest eigenvalue of the matrix
	   * used in the iteration, or the lowest eigvals;
	   * the Ritz vectors have been normalized to length 1 already
	   */

          number_of_eigpairs_to_save = Lanz_M;
          break;
  }
  case 1: case 3:{ 
          /* when looking for those Lanz_Nd eigenvalues of the
	   * original matrix that are closest to folding_point
	   */ 

          number_of_eigpairs_to_save = Lanz_Nd;


          /* using the eigenvectors of the squared-shifted matrix to
	   * obtain the wanted Lanz_Nd eigenvalues of the original matrix;
	   * the Ritz vectors have been normalized to length 1 already
	   */
          for(i=1;i<=Lanz_Nd;i++){
	          for(j=1;j<=n;j++){
		          tmp_nvec[j] = 0.0;
			  l=1;
			  while(l<=degree[j]){
			          tmp_nvec[j] += q[i][NEI[j][l]];
				  l++;
			  }
		  }
		  tmpd = 0.0;
		  tmpd2 = 0.0;
		  for(j=1;j<=n;j++){
			  tmpd += tmp_nvec[j]*q[i][j];
		          tmpd2 += MY_SQR(q[i][j]);
		  }
		  tmp_alpha = alpha[i];
		  alpha[i] = tmpd/tmpd2;
	  }

	  /* arranging the eigenvalues into descending order (and the
	   * eigenvectors into the corresponding order, too)
	   */

	  /* index-sorting the Lanz_Nd eigvals into ___ascending___ order */ 
	  indexx_ov_d_ul(Lanz_Nd,alpha,indx_d_tqli);
          /* we'll need the ___descending___ order: reverting */
          for(i=1;i<=Lanz_Nd;i++){
	          new_indx[i] = indx_d_tqli[Lanz_Nd+1-i];
	  }
	  /* the alpha, beta and erp arrays */
	  for(i=1;i<=Lanz_Nd;i++){ tmp_mvec[i] = alpha[new_indx[i]]; }
	  for(i=1;i<=Lanz_Nd;i++){ alpha[i] = tmp_mvec[i]; }
	  for(i=1;i<=Lanz_Nd;i++){ tmp_mvec[i] = beta[new_indx[i]]; }
	  for(i=1;i<=Lanz_Nd;i++){ beta[i] = tmp_mvec[i]; }
	  /* 
	   *	  for(i=1;i<=Lanz_Nd;i++){ tmp_mvec[i] = erp[new_indx[i]]; }
	   *	  for(i=1;i<=Lanz_Nd;i++){ erp[i] = tmp_mvec[i]; }
	   */
	  /* putting the eigenvectors into the correct order */
	  for(j=1;j<=n;j++){
	          for(i=1;i<=Lanz_Nd;i++){
		          tmp_mvec[i] = q[new_indx[i]][j];
		  }
		  for(i=1;i<=Lanz_Nd;i++){
		          q[i][j] = tmp_mvec[i];
		  }
	  }

          break;
  }
  case 10: case 30:{
          fprintf(stderr,"Sorry, fold_sw=10 and 30 are not finished yet.\n");
	  fprintf(stderr,"Exit / GetEigVals.3.1\n");
	  MY_EXIT;
	  break;
  }
  }
  


  /* 3.2 errors of Ritz pairs 
   * erp_i = | x^T_i A x_i | / | x^T_i x_i | - lambda_i
   * The Ritz vectors have all been normalized to length 1
   * previously
   */
  switch(fold_sw){
  case 0: case 1: case 3: default:{

          /* when eigenpairs of the original matrix are to be returned
	   */
          for(i=1;i<=number_of_eigpairs_to_save;i++){
	          szendvics[i] = 0.0;
		  for(j=1;j<=n;j++){
		          tmp_nvec[j] = 0.0;
			  l=1;
			  while(l<=degree[j]){
			          tmp_nvec[j] += q[i][NEI[j][l]];
				  l++;
			  }
			  szendvics[i] += q[i][j]*tmp_nvec[j];
		  }
		  erp[i] = szendvics[i] - alpha[i];
	  }


/* 
 * writing the largest row sum into this debug file
 */

if(debug[0]==1){
  fprintf(fp[0],"\n# largest row sum\n");
  fprintf(fp[0],"\n");
  tmp_fnvec=vector(1,n);
  for(i=1;i<=n;i++){
    tmp_fnvec[i]=degree[i];
  }
  sort(n,tmp_fnvec);
  fprintf(fp[0],"%f\n",tmp_fnvec[n]);
  free_vector(tmp_fnvec,1,n);
}


/* 
 * ANOTHER OPTION (SWITCHED OFF AT THE MOMENT)
 * writing degrees of vertices and eigenvectors of the largest
 * "number_of_eigpairs_to_save" eigenvalues into data file 
 */

/*
 *if(debug[0]==1){
 *  fprintf(fp[0],"\n# degrees and components of eigenvectors\n");
 *  fprintf(fp[0],"\n# i\tdegree");
 *  for(i=1;i<=number_of_eigpairs_to_save;i++){
 *    fprintf(fp[0],"\tv[%d]",i);
 *  }
 *  fprintf(fp[0],"\n\n");
 *  for(j=1;j<=n;j++){
 *    fprintf(fp[0],"%d\t%ld",j,MS0_Degree[j]);
 *    for(i=1;i<=number_of_eigpairs_to_save;i++){
 *      fprintf(fp[0],"\t%.3E",q[i][j]);
 *    }
 *    fprintf(fp[0],"\n\n");
 *  }
 *}
 */


	  break;
  }
  case 2:{ 

          /* when eigenpairs of the folded matrix are to be returned */
          for(i=1;i<=number_of_eigpairs_to_save;i++){

		  /* storing   A q_i   in    q_{Lanz_M+1} */
		  for(j=1;j<=n;j++){
			  q[Lanz_M+1][j] = 0.0;
			  l=1;
			  while(l<=degree[j]){
			          q[Lanz_M+1][j] += q[i][NEI[j][l]];
				  l++;
			  }
		  }
		  /* storing   A^2 q_i   in   tmp_nvec */
		  for(j=1;j<=n;j++){
		          tmp_nvec[j] = 0.0;
			  l=1;
			  while(l<=degree[j]){
			          tmp_nvec[j] += q[Lanz_M+1][NEI[j][l]];
				  l++;
			  }
		  }
		  /* adding   -2*A*folding_point */
		  for(j=1;j<=n;j++){
		          tmp_nvec[j] -= 2.0*folding_point*q[Lanz_M+1][j];
		  }
		  /* adding   I*folding_point^2 */
		  for(j=1;j<=n;j++){
		          tmp_nvec[j] += MY_SQR(folding_point)*q[i][j];
		  }


		  /* computing    q_i  (A-folding_point*I)^2  q_i */
	          szendvics[i] = 0.0;
		  for(j=1;j<=n;j++){
			  szendvics[i] += q[i][j]*tmp_nvec[j];
		  }
		  erp[i] = szendvics[i] - alpha[i];
	  }
	  break;
  }
  case 10: case 30:{
          fprintf(stderr,"Sorry, fold_sw=10 and 30 are not finished yet.\n");
	  fprintf(stderr,"Exit / GetEigVals.3.2\n");
	  MY_EXIT;
	  break;
  }
  case 4:{
          /* the highest eigvals of the matrix -A are in the array in
	   * descending order
	   * multiplying them by -1 will give the lowest eigvals of A
	   * in ascending order
	   */

          for(i=1;i<=number_of_eigpairs_to_save;i++){
	          alpha[i] *= -1.0;
	  }
  }
  }

  
  /* 3.3 */

  /* return values */
  *return_sum_omega_abs = sum_omega_abs;
  *return_n_of_iter = n_of_iter;
  for(i=1;i<=number_of_eigpairs_to_save;i++){
          return_lambda[i] = alpha[i];
	  return_beta[i] = beta[i];
	  return_error[i] = erp[i];
  }
  /* filling up arrays with 0's */
  while(i<=Lanz_M){
          return_lambda[i] = 0.0;
	  return_beta[i] = 0.0;
	  return_error[i] = 0.0;
	  i++;
  }


/* writing the components of q_1 vs. the degrees of vertices */
/*
 *if(fold_sw==0){
 *  fprintf(fp[6],"\n");
 *  fprintf(fp[6],"# lambda_1 = %.3E\n",return_lambda[1]);
 *  fprintf(fp[6],"# i\tdegree\tq_{1i}\n");
 *  fprintf(fp[6],"\n\n");
 *  for(i=1;i<=n;i++){
 *    fprintf(fp[6],"%ld\t%ld\t%.5E\n",i,degree[i],q[1][i]);
 *  }
 *}
 */



  /* 3.4 */
  for(i=0;i<=7;i++){
          if(debug[i]==1){
	          fflush(fp[i]);
	          fclose(fp[i]);
	  }
  }




  /* 3.5 */
  free_ivector(debug,0,7);
  free_dmatrix(q,1,Lanz_M+1,1,n);
  free_dvector(beta,1,Lanz_M);
  free_dvector(alpha,1,Lanz_M);
  free_dvector(d_tqli,1,Lanz_M);
  free_dvector(e_tqli,1,Lanz_M);
  free_dmatrix(a_tqli,1,Lanz_M,1,Lanz_M);
  free_lvector(indx_d_tqli,1,Lanz_M);
  free_dvector(erp,1,Lanz_M);
  free_dvector(szendvics,1,Lanz_M);
  free_dvector(tmp_nvec,1,n);
  free_dvector(tmp_mvec,1,Lanz_M);
  free_lvector(new_indx,1,Lanz_M);
  free_dvector(ritz_vec_sqr,1,Lanz_M);
  free_dmatrix(omega,1,Lanz_M+1,1,Lanz_M+1);
}

/*------------------------------------------------*/

void SortedEigVals_NumRec( float **a, int n, float *eigval ){
  /* sorted eigenvalues of the  n x n   real symmetric matrix a 
   * using routines from "Numerical Recipes"
   */

  int i;
  float *d,*e;

  d=vector(1,n);
  e=vector(1,n);
  tred2_ne(a,n,d,e);
  tqli_ne_il(d,e,n,tqli_ItMax);

  sort( (unsigned long)n, d );
  for(i=1;i<=n;i++){ eigval[i]=d[i]; }
  free_vector(d,1,n);
  free_vector(e,1,n);
}

/******************************************/

void MS0_Init_Outside(){

  int ii;
  unsigned long i;


  /* global variables */
  MS0_FirstSeed=RndSeed;
  MS0_LastSeed=RndSeed+NofSeeds-1;

  switch(Model){
  case 0: default:{
          MS0_nVertex =   (unsigned long)10000*ER_nVertex_div10000
	                + (unsigned long)ER_nVertex_mod10000;
	  MS0_nEdge =     (unsigned long)10000*ER_nEdge_div10000 
	                + (unsigned long)ER_nEdge_mod10000;
	  break;
  }
  case 1:{
          BA_T =   (unsigned long)10000*BA_Tdiv10000 
	         + (unsigned long)BA_Tmod10000;
	  MS0_nVertex = (unsigned long)BA_M0 + BA_T;
	  MS0_nEdge = (unsigned long)BA_M * BA_T;
	  break;
  }
  case 2: case 20:{
          UC_NTryLimit =   (unsigned long)10000*UC_NTryLimit_div10000 
	                 + (unsigned long)UC_NTryLimit_mod10000;
          MS0_nVertex =   (unsigned long)10000*UC_nVertex_div10000 
	                + (unsigned long)UC_nVertex_mod10000;
	  break;
  }
  case 3:{
          MS0_nVertex =   (unsigned long)10000*SW_nVertex_div10000
	                + (unsigned long)SW_nVertex_mod10000;
	  MS0_nEdge = MS0_nVertex * SW_k/2;
	  break;
  }
  }


  /* for MainSwitch=100 and 101 */
  MS0_RowProd_Stat=dvector(1,MS0_nVertex);
  MS0_RowProd_StatCounter=dvector(1,MS0_nVertex);
  for(i=1;i<=MS0_nVertex;i++){
          MS0_RowProd_Stat[i]=0.0;
          MS0_RowProd_StatCounter[i]=0.0;
  }



  /* specific init for different folding schemes */

  /* each memory allocation line needed for every scheme */
  MS0_NofIter=ivector(MS0_FirstSeed,MS0_LastSeed);
  MS0_Lambda=dmatrix(MS0_FirstSeed,MS0_LastSeed,1,Lanz_M);
  MS0_Beta=dmatrix(MS0_FirstSeed,MS0_LastSeed,1,Lanz_M);
  MS0_Error=dmatrix(MS0_FirstSeed,MS0_LastSeed,1,Lanz_M);
  MS0_SumOmegaAbs=dvector(MS0_FirstSeed,MS0_LastSeed);


  MS0_Lambda0num_I=15;
  MS0_Lambda0_I=dvector(1,MS0_Lambda0num_I);
  for(ii=1;ii<=MS0_Lambda0num_I;ii++){
          MS0_Lambda0_I[ii]=-5.0+1.0*(ii-1.0);
  }
  MS0_NofIter_I=imatrix(MS0_FirstSeed,MS0_LastSeed,1,MS0_Lambda0num_I);
  MS0_Lambda_I=d3tensor(MS0_FirstSeed,MS0_LastSeed,
			1,MS0_Lambda0num_I,1,Lanz_M
			);
  MS0_Beta_I=d3tensor(MS0_FirstSeed,MS0_LastSeed,
		      1,MS0_Lambda0num_I,1,Lanz_M
		      );
  MS0_Error_I=d3tensor(MS0_FirstSeed,MS0_LastSeed,
		       1,MS0_Lambda0num_I,1,Lanz_M
		       );
  MS0_SumOmegaAbs_I=d3tensor(MS0_FirstSeed,MS0_LastSeed,
			     1,MS0_Lambda0num_I,1,Lanz_M
			     );

  /* NOSC: # of small clusters */
  MS0_NOSC=lvector(1,4);
}

/*-------------------------------------------*/

void MS0_Init_Inside(){

  MS0_Degree=lvector(1,MS0_nVertex);
}

/*-----------------------------------------------------*/

void MS0_Run( int main_switch, int seed, int fold_switch, int stretch,
	      int *n_of_iter, double *lambda,
	      double *beta, double *error, double *sum_omega_abs, 
	      unsigned long *degree,
	      int lambda_0_num_I, double *lambda_0_I,
	      int *n_of_iter_I, double **lambda_I,
	      double **beta_I, double **error_I, 
	      double **sum_omega_abs_I,
	      double *rowprod_stat, 
	      double *rowprod_statcounter,
	      unsigned long *nosc
	    ){

  double *ipr_dummy;
  /* ipr_dummy: dummy variable */



  /* generating matrix */
  GenerateMatrix( seed, Model, 
		  MS0_nVertex, MS0_nEdge, 
		  BA_M0, BA_M, BA_T,
		  UC_Gamma, MS0_nVertex, &MS0_UC_nEdge,
		  UC_minDegree, UC_pN, UC_NTryLimit,
		  MS0_nVertex, SW_k, SW_p, degree
                );


  switch(main_switch){
  case 0:{ /* computing the number of small clusters 
	    * and eigenvalues (largest or internal) of the matrix
	    */

          /* SPECIAL CASE:
           * counting and saving the number of small clusters,
	   * and exiting afterwards 
	   */
          if(Lanz_NOSC==1){
	          MS0_Cnosc( nosc );
		  exit(-2);
	  }


          /* EIGENVALUES: 
	   * what do we want? largest eigval or internal specdens? 
	   */
          switch(fold_switch){
	  case 0: default:{ /* searching for largest eigenvalue */

	          GetEigVals( MS0_nVertex, 0, 0.0, 0, 0.0,
			      degree, n_of_iter, lambda, beta, error,
			      sum_omega_abs, ipr_dummy
			    );
		  break;
	  }
	  case 1: case 10:{ /* searching for internal eigvals */
	          double shift,lambda_1_of_squared,folding_point;
		  int i,tmpi,fold_sw_STRETCH_ON,fold_sw_STRETCH_OFF;
		  
		  /* get the Lanz_Nd eigenvalues closest to every
		   * lambda_0 (and their eigenvectors)  
		   */
		  for(i=1;i<=lambda_0_num_I;i++){

fprintf(stderr,"lambda_0_I=%.3E\t",lambda_0_I[i]);system("date");



                          /* 0. 
			   * set point of folding, 
			   * and the ON/OFF values for stretching
			   */
                          switch(fold_switch){
			  case 1: default:{
			          folding_point=lambda_0_I[i];
				  fold_sw_STRETCH_OFF=1;
				  fold_sw_STRETCH_ON=3;
				  break;
			  }
			  case 10:{
			          folding_point=lambda_0_I[i]+Lanz_Phi;
				  fold_sw_STRETCH_OFF=10;
				  fold_sw_STRETCH_ON=30;
				  break;
			  }
			  }



			  /* 1.
			   * get the largest eigval of the squared matrix
			   * (this value will be at the lower end of the
			   * spectrum after substracting and shifting)
			   */
			  GetEigVals( MS0_nVertex, 2, 0.0, 0,
				      folding_point,
				      degree, &tmpi, lambda,
				      beta, error, 
				      sum_omega_abs, ipr_dummy
				    );
			  lambda_1_of_squared = lambda[1];


fprintf(stderr,"\tlambda_1=%.3E\t",lambda[1]);system("date");



                          /* 2.
			   * get the eigvals around lambda_0
			   * for this, use the matrix
			   *        shift*I - (A-folding_point*I)^2 ,
			   * where 
			   *        shift = lambda_1_of_squared / 2
			   *
			   * if Lanz_Fold=1:
			   * we do this to obtain a matrix whose
			   * spectrum is between  
			   * -x and x, and where the highest eigvals (those
			   * closest to x) correspond to the
			   * eigvals of the original matrix that are closest 
			   * to lambda_0 
			   *
			   * if Lanz_Fold=10:
			   * folding similarly, but not the top
			   * eigvals of the transformed matrix are kept
			   *
			   */
                          shift = 0.5*lambda_1_of_squared;
			  
			  if(stretch>1){ /* if extra stretch is needed */
			          GetEigVals( MS0_nVertex, fold_sw_STRETCH_ON, 
					      shift, stretch,
					      folding_point, degree,
					      &(n_of_iter_I[i]),  
					      lambda_I[i], beta_I[i], 
					      error_I[i],
					      sum_omega_abs_I[i],
					      ipr_dummy
					    );
			  }
			  else{ /* no stretch */
			          GetEigVals( MS0_nVertex, fold_sw_STRETCH_OFF,
					      shift, 0,
					      folding_point, degree,
					      &(n_of_iter_I[i]),  
					      lambda_I[i], beta_I[i], 
					      error_I[i],
					      sum_omega_abs_I[i],
					      ipr_dummy
					    );
			  }
		  }
		  break;
 	  }
	  }

  	  break;
  }
  case 101:{ /* computing product matrix of row vectors */

	  unsigned long i,j,k,i_indx,j_indx,bin_indx,nb;
	  double tmp_rowprod;

	  for(i=1;i<=MS0_nVertex;i++){
	          for(j=i;j<=MS0_nVertex;j++){
		          if(j==i){
			          tmp_rowprod=(double)degree[i];
			  }
			  else{
			          tmp_rowprod=0.0;
				  i_indx=j_indx=1;
				  while(  (i_indx<=degree[i])
					&&(j_indx<=degree[j])
				       ){
				          if(NEI[i][i_indx]==NEI[j][j_indx]){
					          tmp_rowprod+=1.0;
					  }
					  k = MY_MIN( NEI[i][i_indx],
						      NEI[j][j_indx]
						    );
					  if(k==NEI[i][i_indx]){i_indx++;}
					  if(k==NEI[j][j_indx]){j_indx++;}
				  }
			  }


			  nb=ARP_NofBins;
			  if(j==nb){ 
			          bin_indx=nb; 
			  }
			  else{
			          bin_indx = 1 + (int)floor( MY_ABS(log((double)j/(double)i))
							    /(log(nb)/(double)nb)
							   );
			  }
			  rowprod_stat[bin_indx] += tmp_rowprod;
			  rowprod_statcounter[bin_indx] += 1.0;
		  }
	  }
          break;
  }
  }
}

/*--------------------------------------*/

void MS0_Clean_Inside(){

  unsigned long i,*nei_ncl,*nei_nch;
  nei_ncl=lvector(1,MS0_nVertex);
  nei_nch=lvector(1,MS0_nVertex);
  
  for(i=1;i<=MS0_nVertex;i++){
          nei_ncl[i] = 1;
          if(MS0_Degree[i]==0){ nei_nch[i] = 1; }
	  else{ nei_nch[i] = MS0_Degree[i]; }
  }
  free_lmatrix_vrl(NEI,1,MS0_nVertex,nei_ncl,nei_nch);
  free_lvector(MS0_Degree,1,MS0_nVertex);
  free_lvector(nei_ncl,1,MS0_nVertex);
  free_lvector(nei_nch,1,MS0_nVertex);
}

/*--------------------------------------*/

void MS0_EvalSave_AtEnd( int main_switch, int fold, int *n_of_iter,
			 double **lambda, 
			 double **beta, double **error, 
			 double *sum_omega_abs,
			 int lambda_0_num_I, double *lambda_0_I,
			 int **n_of_iter_I, double ***lambda_I,
			 double ***beta_I, double ***error_I,
			 double ***sum_omega_abs_I,
			 double *rowprod_stat, double *rowprod_statcounter
		       ){

  int i,seed,j,k;
  double e,sigma;
  FILE *fp,*fp2;
  char fn[MY_STRLEN],fn2[MY_STRLEN];


  switch(main_switch){ 
  case 0: default:{ /* looking for some eigvals */


      switch(fold){
      case 0:{ /* when looking for largest eigval */
    
          switch(Model){
	  case 0: default: {
	          sprintf(fn,"%s_ER_lambda1_nVertex%ld_nEdge%ld_seeds%dto%d_",
			  OFN,MS0_nVertex,MS0_nEdge,MS0_FirstSeed,MS0_LastSeed);
		  break;
	  }
	  case 1: {
	          sprintf(fn,"%s_BA_lambda1_m=m0=%d_t%ld_seeds%dto%d_",
			  OFN,BA_M,BA_T,MS0_FirstSeed,MS0_LastSeed);
		  break;
	  }
	  case 2:{
	          sprintf(fn,"%s_UC2_lambda1_nVertex=%ld_Gamma=%g_seeds%dto%d_",
			  OFN,MS0_nVertex,UC_Gamma,MS0_FirstSeed,MS0_LastSeed);
		  break;
	  }
	  case 20:{
             sprintf(fn,"%s_UC20_lambda1_nVertex=%ld_Gamma=%g_pN=%g_seeds%dto%d_",
			  OFN,MS0_nVertex,UC_Gamma,UC_pN,
			  MS0_FirstSeed,MS0_LastSeed);
		  break;
	  }
	  }
	  fp=fopen(fn,"w");
	  if(!fp){
	          fprintf(stderr,"ERROR: couldn't open the file\n  %s\nfor writing",fn);
		  MY_EXIT;
	  }



	  fprintf(fp,"\n");
	  fprintf(fp,"# largest eigval of graph\n");
	  fprintf(fp,"# seed:\t");
          for(seed=MS0_FirstSeed;seed<=MS0_LastSeed;seed++){ 
	          fprintf(fp,"\t%d\t",seed);
	  }
	  fprintf(fp,"\tE\t\tsigma\n");


	  fprintf(fp,"\nlambda_1");
	  e=0.0;
	  for(seed=MS0_FirstSeed;seed<=MS0_LastSeed;seed++){ 	  
	          fprintf(fp,"\t%8.4E",lambda[seed][1]);
		  e+=lambda[seed][1];
	  }
	  e/=(double)(MS0_LastSeed-MS0_FirstSeed+1);
	  sigma=0.0;
	  for(seed=MS0_FirstSeed;seed<=MS0_LastSeed;seed++){ 	  
		  sigma+=MY_SQR(lambda[seed][1]-e);
	  }
	  sigma/=(double)(MS0_LastSeed-MS0_FirstSeed+1);
	  sigma=sqrt(sigma);
	  fprintf(fp,"\t%8.4E\t%8.4E",e,sigma);


	  fprintf(fp,"\nerp\t");
	  e=0.0;
	  for(seed=MS0_FirstSeed;seed<=MS0_LastSeed;seed++){ 	  
	          fprintf(fp,"\t%8.4E",error[seed][1]);
		  e+=error[seed][1];
	  }
	  e/=(double)(MS0_LastSeed-MS0_FirstSeed+1);
	  sigma=0.0;
	  for(seed=MS0_FirstSeed;seed<=MS0_LastSeed;seed++){ 	  
		  sigma+=MY_SQR(error[seed][1]);
	  }
	  sigma/=(double)(MS0_LastSeed-MS0_FirstSeed+1);
	  sigma=sqrt(sigma);
	  fprintf(fp,"\t%8.4E\t%8.4E",e,sigma);


	  fprintf(fp,"\nsum |w|\t");
	  e=0.0;
	  for(seed=MS0_FirstSeed;seed<=MS0_LastSeed;seed++){ 	  
	          fprintf(fp,"\t%8.4E",sum_omega_abs[seed]);
		  e+=sum_omega_abs[seed];
	  }
	  e/=(double)(MS0_LastSeed-MS0_FirstSeed+1);
	  sigma=0.0;
	  for(seed=MS0_FirstSeed;seed<=MS0_LastSeed;seed++){ 	  
		  sigma+=MY_SQR(sum_omega_abs[seed]);
	  }
	  sigma/=(double)(MS0_LastSeed-MS0_FirstSeed+1);
	  sigma=sqrt(sigma);
	  fprintf(fp,"\t%8.4E\t%8.4E\n",e,sigma);
	  

	  fprintf(fp,"\n\n");
	  fflush(fp);
	  fclose(fp);
          break;
      }
      case 1:{ /* when looking for inside spectral density */

          double **distance;
	  distance = dmatrix(MS0_FirstSeed,MS0_LastSeed,1,lambda_0_num_I);


          switch(Model){
	  case 0: default: {
	          sprintf(fn,"%s_ER_internal_nVertex%ld_nEdge%ld_seeds%dto%d_",
			  OFN,MS0_nVertex,MS0_nEdge,MS0_FirstSeed,MS0_LastSeed);
		  break;
	  }
	  case 1: {
	          sprintf(fn,"%s_BA_internal_m=m0=%d_t%ld_seeds%dto%d_",
			  OFN,BA_M,BA_T,MS0_FirstSeed,MS0_LastSeed);
		  break;
	  }
	  case 2:{
	          sprintf(fn,"%s_UC2_internal_nVertex=%ld_Gamma=%g_seeds%dto%d_",
			  OFN,MS0_nVertex,UC_Gamma,MS0_FirstSeed,MS0_LastSeed);
		  break;
	  }
	  case 20:{
	    sprintf(fn,"%s_UC20_internal_nVertex=%ld_Gamma=%g_pN=%g_seeds%dto%d_",
			  OFN,MS0_nVertex,UC_Gamma,UC_pN,
			  MS0_FirstSeed,MS0_LastSeed);
		  break;
	  }
	  }
	  sprintf(fn2,"%s_ALL_",fn);

	  fp=fopen(fn,"w");
	  if(!fp){
	          fprintf(stderr,"ERROR: couldn't open the file\n  %s\nfor writing",fn);
		  MY_EXIT;
	  }
	  fp2=fopen(fn2,"w");
	  if(!fp2){
	          fprintf(stderr,"ERROR: couldn't open the file\n  %s\nfor writing",fn2);
		  MY_EXIT;
	  }



	  /* first data file: summary */
	  fprintf(fp,"\n");
	  fprintf(fp,"\n");
	  fprintf(fp,"# SUMMARY FILE\n");
	  fprintf(fp,"\n");
	  fprintf(fp,"# Average distance between an eigval and its nth neighbor\n");
	  fprintf(fp,"# [av. dist. between the (a)th to the left and the (b)th to the right\n");
	  fprintf(fp,"#  from the base point lambda_0, where a+b=n+1]\n");
	  fprintf(fp,"\n");
	  fprintf(fp,"# lambda_0\tn\te(distance)\tsigma(dist.)\tnum. of seeds\n");
	  for(i=1;i<=lambda_0_num_I;i++){
	          fprintf(fp,"%6.3E\t%d\t",lambda_0_I[i],Lanz_Nd-1);
		  e = 0.0;
		  for(seed=MS0_FirstSeed;seed<=MS0_LastSeed;seed++){
		          /* !!! eigvals are stored in descending order */
		          distance[seed][i] =   lambda_I[seed][i][1]
			                      - lambda_I[seed][i][Lanz_Nd];
			  e+=distance[seed][i];
		  }
		  e /= (double)(MS0_LastSeed-MS0_FirstSeed+1);
		  sigma = 0.0;
		  for(seed=MS0_FirstSeed;seed<=MS0_LastSeed;seed++){
			  sigma+=MY_SQR(e-distance[seed][i]);
		  }
		  sigma /= (double)(MS0_LastSeed-MS0_FirstSeed+1);
		  sigma = sqrt(sigma);
		  fprintf(fp,"%.5E\t%.5E\t%d\n",
			  e,sigma,MS0_LastSeed-MS0_FirstSeed+1
			 );
	  }
	  


          /* 2nd data file: detailed data */
	  fprintf(fp2,"\n");
	  fprintf(fp2,"\n");
	  fprintf(fp2,"# DETAILED DATA\n");
	  fprintf(fp2,"\n");
	  fprintf(fp2,"# Eigvals around the base point lambda_0\n"); 
	  fprintf(fp2,"# for matrices generated with different rnd seeds.\n");
	  fprintf(fp2,"\n");
	  fprintf(fp2,"# seed\tlambda_0\tindex\teigval\t\tit.beta\t\tnew erp\n");
	  for(seed=MS0_FirstSeed;seed<=MS0_LastSeed;seed++){
	          fprintf(fp2,"\n");
	          for(i=1;i<=lambda_0_num_I;i++){
		          fprintf(fp2,"\n");
		          for(j=1;j<=Lanz_Nd;j++){
			         fprintf(fp2,"%d\t%.3E\t%d\t%.3E\t%.3E\t%.3E\n",
					  seed,lambda_0_I[i],j,
					  lambda_I[seed][i][j],
					  beta_I[seed][i][j],
					  error_I[seed][i][j]
					 );
			  }
		  }
	  }


	  free_dmatrix(distance,MS0_FirstSeed,MS0_LastSeed,1,lambda_0_num_I);
	  fflush(fp);
	  fclose(fp);
	  fflush(fp2);
	  fclose(fp2);
	  break;
      }
      }
      break;
  }
  case 101:{ /* taking the average of av. row product values */
    
        unsigned long indx,nb;

	switch(Model){
	case 0: default: {
                sprintf(fn,"%s_ER_arp_nVertex%ld_nEdge%ld_seeds%dto%d_",
			OFN,MS0_nVertex,MS0_nEdge,
			MS0_FirstSeed,MS0_LastSeed
		       );
		break;
	}
	case 1: {
                sprintf(fn,"%s_BA_arp_m=m0=%d_t%ld_seed%dto%d_",
			OFN,BA_M,BA_T,
			MS0_FirstSeed,MS0_LastSeed
			);
		break;
	}
	case 2:{
                sprintf(fn,"%s_UC2_arp_nVertex=%ld_Gamma=%g_seed%dto%d_",
			OFN,MS0_nVertex,UC_Gamma,
			MS0_FirstSeed,MS0_LastSeed
			);
		break;
	}
	case 20:{
                sprintf(fn,"%s_UC20_arp_nVertex=%ld_Gamma=%g_pN=%g_seed%dto%d_",
			OFN,MS0_nVertex,UC_Gamma,UC_pN,
			MS0_FirstSeed,MS0_LastSeed
			);
		break;
	}
	}
	fp=fopen(fn,"w");
	if(!fp){
                fprintf(stderr,"ERROR: couldn't open the file\n  %s\nfor writing",fn);
		MY_EXIT;
	}
	
	
	
	nb=MS0_nVertex;
	fprintf(fp,"\n");
	fprintf(fp,"# bin center\taverage f[i_1][i_2]\n");
	fprintf(fp,"\n\n");
	for(indx=1;indx<=nb;indx++){
	        fprintf(fp,"%.3E\t%.3E\n",
			exp((indx-0.5)*(log(nb)/(double)nb)),
			rowprod_stat[indx] / rowprod_statcounter[indx]
			);
	}
	fflush(fp);
	fclose(fp);


	break;
  }
  }
}


/*--------------------------------------*/

void MS0_Cnosc( unsigned long *nosc ){
  /* counting the number of small isolated clusters */

  unsigned long i,j,k,l;

  /* clusters of size 1 (i.e., isolated vertices */
  nosc[1]=0;
  for(i=1;i<=MS0_nVertex;i++){
          if(MS0_Degree[i]==0){ nosc[1]++; }
  }

  /* isolated clusters of size 2 */
  nosc[2]=0;
  for(i=1;i<=MS0_nVertex;i++){
          if((MS0_Degree[i]==1)&&(MS0_Degree[NEI[i][1]]==1)){
	          nosc[2]++;
	  }
  }
  nosc[2]/=2;

  /* isolated clusters with 3 vertices and 2 edges */
  nosc[3]=0;
  for(i=1;i<=MS0_nVertex;i++){
          if(MS0_Degree[i]==1){
	          j=NEI[i][1];
		  if(MS0_Degree[j]==2){
			  if((k=NEI[j][1])==i){
			          k=NEI[j][2];
			  }
			  if(MS0_Degree[k]==1){
			          nosc[3]++;
			  }
		  }
	  }
	  else if(MS0_Degree[i]==2){
	          j=NEI[i][1];
	          k=NEI[i][2];
		  if((MS0_Degree[j]==1)&&(MS0_Degree[k]==1)){
		          nosc[3]++;
		  }
	  }
  }
  nosc[3]/=3;

  /* isolated clusters with 3 vertices and 3 edges */
  nosc[4]=0;
  for(i=1;i<=MS0_nVertex;i++){
          if(MS0_Degree[i]==2){
	          j=NEI[i][1];
	          k=NEI[i][2];
		  if((MS0_Degree[j]==2)&&(MS0_Degree[k]==2)){		  
		          if((l=NEI[j][1])==i){		          
			          l=NEI[j][2];
			  }
			  if(l==k){
			          nosc[4]++;
			  }
		  }      
	  }
  }
  nosc[4]/=3;


  /* data output */
  fprintf(stdout,"\n\n");
  fprintf(stdout,"# Number of small isolated clusters.\n");
  fprintf(stdout,"# 1vertex, 2vertices, 3vertices+2edges, 3vertices+3edges\n");
  fprintf(stdout,"%ld,%ld,%ld,%ld\n",nosc[1],nosc[2],nosc[3],nosc[4]);
  fprintf(stdout,"\n");
}

/*************************************************/

void MS1_Init(){

  BA_T =   (unsigned long)10000*BA_Tdiv10000 
         + (unsigned long)BA_Tmod10000;
  UC_NTryLimit =   (unsigned long)10000*UC_NTryLimit_div10000 
                 + (unsigned long)UC_NTryLimit_mod10000;

  switch(Model){
  case 0: default:{
          MS1_nVertex =   (unsigned long)10000*ER_nVertex_div10000
	                + (unsigned long)ER_nVertex_mod10000;
	  MS1_nEdge =     (unsigned long)10000*ER_nEdge_div10000 
	                + (unsigned long)ER_nEdge_mod10000;
	  break;
  }
  case 1:{
	  MS1_nVertex = (unsigned long)BA_M0 + BA_T;
	  MS1_nEdge = (unsigned long)BA_M * BA_T;
	  break;
  }
  case 2: case 20:{
          MS1_nVertex =   (unsigned long)10000*UC_nVertex_div10000 
	                + (unsigned long)UC_nVertex_mod10000;
	  break;
  }
  case 3:{
          MS1_nVertex =   (unsigned long)10000*SW_nVertex_div10000
	                + (unsigned long)SW_nVertex_mod10000;
	  MS1_nEdge = MS1_nVertex * SW_k/2;
	  break;
  }
  }

  MS1_Degree=lvector(1,MS1_nVertex);
}

/*--------------------------------------*/

void MS1_Run(){

  GenerateMatrix( RndSeed, Model, 
		  MS1_nVertex, MS1_nEdge, 
		  BA_M0, BA_M, BA_T,
		  UC_Gamma, MS0_nVertex, &MS1_UC_nEdge,
		  UC_minDegree, UC_pN, UC_NTryLimit,
		  MS1_nVertex, SW_k, SW_p,
		  MS1_Degree
                );
  MS1_Save();
}

/*--------------------------------------*/

void MS1_Save(){
  /*******************************************
   * saving matrix stored in the arrays NEI and MS1_Degree in row
   * list format:
   *
   **********************************************
   * DATA FILE FORMAT
   *
   * info                      how many bytes
   * -----------------------------------------
   * number of vertices        4 
   *     nVertex = ((b1*256+b2)*256+b3)*256+b4
   * list of degrees           4*nVertex
   * list of non-zero entries  4*nEdge*2
   *
   * the parameter file is concatenated to the end of the data file
   * in ascii format
   **********************************************/
  
  unsigned long i,j;
  FILE *fp;
  char fn[MY_STRLEN],shell_command[MY_STRLEN];
  int int1,int2,int3,int4;


  switch(Model){
  case 0: default:{
            sprintf(fn,"rndma._ER_nVertex=%ld_nEdge=%ld_RndSeed=%ld_.dat",
		    MS1_nVertex,MS1_nEdge,RndSeed
		    );
	    break;
  }
  case 1:{
            sprintf(fn,"rndma._BA_M=%ld_M0=%ld_T=%ld_RndSeed=%ld_.dat",
		    BA_M,BA_M0,BA_T,RndSeed);
	    break;
  }
  case 2: case 20:{
            sprintf(fn,"rndma._UC_N=%ld_Gamma=%f_kMin=%ld_RndSeed=%ld_.dat",
		    (unsigned long)(10000*UC_nVertex_div10000 + UC_nVertex_mod10000),
		    UC_Gamma,UC_minDegree,RndSeed);
            break;
  }
  }


  fp=fopen(fn,"w");
  if(!fp){
          fprintf(stderr,"ERROR: couldn't open file for writing: %s\n",fn);
	  MY_EXIT;
  }

  /* number of vertices */
  int1 = MS1_nVertex % 256;
  int2 = (MS1_nVertex / 256) % 256;
  int3 = ((MS1_nVertex / 256) / 256) % 256;
  int4 = ((MS1_nVertex / 256) / 256) / 256;
  fprintf(fp,"%c%c%c%c",int1,int2,int3,int4);

  /* list of degrees */
  for(i=1;i<=MS1_nVertex;i++){
          int1 = MS1_Degree[i] % 256;
	  int2 = (MS1_Degree[i] / 256) % 256;
	  int3 = ((MS1_Degree[i] / 256) / 256) % 256;
	  int4 = ((MS1_Degree[i] / 256) / 256) / 256;
	  fprintf(fp,"%c%c%c%c",int1,int2,int3,int4);
  }

  /* list of non-zero entries */
  for(i=1;i<=MS1_nVertex;i++){
          for(j=1;j<=MS1_Degree[i];j++){
	              int1 = NEI[i][j] % 256;
		      int2 = (NEI[i][j] / 256) % 256;
		      int3 = ((NEI[i][j] / 256) / 256) % 256;
		      int4 = ((NEI[i][j] / 256) / 256) / 256;
		      fprintf(fp,"%c%c%c%c",int1,int2,int3,int4);
	  }
  }
  fflush(fp);
  fclose(fp);

  /* putting parameter file at the end */
  sprintf(shell_command,"echo ' ' >> %s; cat %s >> %s",fn,IFN,fn);
  system(shell_command);
}

/*========================================*/

void MS2_InitAndReadMatrix(){
  /* read matrix in row list format */
  FILE *fp;
  char fn[MY_STRLEN],c0,c1,c2,c3;
  unsigned long i,j,n,*nei_ncl,*nei_nch;
  int i0,i1,i2,i3;

  
  /* 1.
   * init 
   */
  fprintf(stderr,"\nName of data file: ");
  fscanf(stdin,"%s",fn);
  fprintf(stderr,"\n");
  fp=fopen(fn,"r");
  if(!fp){
          fprintf(stderr,"ERROR: couldn't open file for reading: %s",fn);
	  MY_EXIT;
  }


  /* 2. 
   * reading data & memory allocation 
   */

  /* 2.1 */
  fscanf(fp,"%c%c%c%c",&c0,&c1,&c2,&c3);
  if((i0=(int)c0)<0){i0+=256;}
  if((i1=(int)c1)<0){i1+=256;}
  if((i2=(int)c2)<0){i2+=256;}
  if((i3=(int)c3)<0){i3+=256;}
  MS2_nVertex = (unsigned long)(((i3*256+i2)*256+i1)*256+i0);
  n = MS2_nVertex;

/*
 *fprintf(stderr,"%c\n%c\n%c\n%c\n",c0,c1,c2,c3);
 *fprintf(stderr,"\n%ld\n%ld\n%ld\n%ld\n\nMS2_nVertex = %ld\n\n",i0,i1,i2,i3,MS2_nVertex);
 *fflush(stderr);
 *exit(-1);
 */

  MS2_Degree=lvector(1,n);
  nei_ncl=lvector(1,n);
  nei_nch=lvector(1,n);

  /* 2.2 */
  for(i=1;i<=n;i++){
          fscanf(fp,"%c%c%c%c",&c0,&c1,&c2,&c3);
	  if((i0=(int)c0)<0){i0+=256;}
	  if((i1=(int)c1)<0){i1+=256;}
	  if((i2=(int)c2)<0){i2+=256;}
	  if((i3=(int)c3)<0){i3+=256;}
	  MS2_Degree[i] = (unsigned long)(((i3*256+i2)*256+i1)*256+i0);
	  nei_ncl[i] = 1;
	  if(MS2_Degree[i]==0){ nei_nch[i]=1; }
	  else{ nei_nch[i] = MS2_Degree[i]; }
  }
  NEI=lmatrix_vrl(1,n,nei_ncl,nei_nch);


  /* 2.3 */
  for(i=1;i<=n;i++){
          for(j=1;j<=MS2_Degree[i];j++){
	          fscanf(fp,"%c%c%c%c",&c0,&c1,&c2,&c3);	          
		  if((i0=(int)c0)<0){i0+=256;}
		  if((i1=(int)c1)<0){i1+=256;}
		  if((i2=(int)c2)<0){i2+=256;}
		  if((i3=(int)c3)<0){i3+=256;}
		  NEI[i][j] = (unsigned long)(((i3*256+i2)*256+i1)*256+i0);
	  }
  }


  fprintf(stderr,"Finished reading file: %s\n",fn);
  fclose(fp);
  free_lvector(nei_ncl,1,n);
  free_lvector(nei_nch,1,n);
}

/*-------------------------------------------*/

void MS2_Run(){}

/*--------------------------------------*/

void MS2_EvalAndSave(){}

/*********************************************/

void MS200_Init(){

  /* global variables */
  switch(Model){
  case 0: default:{
          MS200_nVertex =   (unsigned long)10000*ER_nVertex_div10000
	                  + (unsigned long)ER_nVertex_mod10000;
	  MS200_nEdge =     (unsigned long)10000*ER_nEdge_div10000 
	                  + (unsigned long)ER_nEdge_mod10000;
	  break;
  }
  case 1:{
          BA_T =   (unsigned long)10000*BA_Tdiv10000 
	         + (unsigned long)BA_Tmod10000;
	  MS200_nVertex = (unsigned long)BA_M0 + BA_T;
	  MS200_nEdge = (unsigned long)BA_M * BA_T;
	  break;
  }
  case 2: case 20:{
          UC_NTryLimit = 10000;
          MS200_nVertex =   (unsigned long)10000*UC_nVertex_div10000 
	                  + (unsigned long)UC_nVertex_mod10000;
	  break;
  }
  case 3:{
          MS200_nVertex =   (unsigned long)10000*SW_nVertex_div10000
	                  + (unsigned long)SW_nVertex_mod10000;
	  MS200_nEdge = MS200_nVertex * SW_k/2;
	  break;
  }
  }


  MS200_Lambda=vector(1,MS200_nVertex);
  MS200_Degree=lvector(1,MS200_nVertex);
}

/*--------------------------------------*/

void MS200_Run( int seed, unsigned long *degree ){

  float **fa;
  unsigned long i,j,k;


  /* generating matrix */
  GenerateMatrix( seed, Model, 
		  MS0_nVertex, MS0_nEdge, 
		  BA_M0, BA_M, BA_T,
		  UC_Gamma, MS0_nVertex, &MS0_UC_nEdge,
		  UC_minDegree, UC_pN, UC_NTryLimit,
		  MS200_nVertex, SW_k, SW_p,
		  degree
                );

  /* eigvals */
  fa=matrix(1,MS200_nVertex,1,MS200_nVertex);
  for(i=1;i<=MS200_nVertex;i++){
          for(j=1;j<=MS200_nVertex;j++){
	          fa[i][j]=0.0;
	  }
  }
  for(i=1;i<=MS200_nVertex;i++){
          j=1;
	  while(j<=MS200_Degree[i]){
	          fa[i][NEI[i][j]]=1.0;
	          j++;
	  }
  }
  SortedEigVals_NumRec( fa, (int)MS200_nVertex, MS200_Lambda );

  free_matrix(fa,1,MS200_nVertex,1,MS200_nVertex);
}

/*--------------------------------------*/

void MS200_EvalAndSave(){

  FILE *fp;
  char fn[MY_STRLEN];
  int i;

  sprintf(fn,"a.dat");
  fp=fopen(fn,"w");
  if(!fp){
          fprintf(stderr,"ERROR: couldn't open file for writing: %s\n",fn);
	  MY_EXIT;
  }

  fprintf(fp,"\n# index\t eigval\n\n");
  for(i=1;i<=MS200_nVertex;i++){
          fprintf(fp,"%d\t%.4E\n",i,MS200_Lambda[MS200_nVertex+1-i]);
  }
  fflush(fp);
  fclose(fp);
}

/*********************************************/

void MS20_Init_Outside(){

  int ii;
  unsigned long i;
  char fn[MY_STRLEN];



  /* global variables */
  /* rnd seeds */
  MS20_FirstSeed=RndSeed;
  MS20_LastSeed=RndSeed+NofSeeds-1;
  /* values of BA_t (the parameter t in the BA model) */
  MS20_Nof_BAT=1; /* number of different values */
  MS20_BAT=lvector(1,MS20_Nof_BAT);
  MS20_BAT[1]=300000;
  MS20_Lambda1=dmatrix(MS20_FirstSeed,MS20_LastSeed,1,MS20_Nof_BAT);
  MS20_Lambda2=dmatrix(MS20_FirstSeed,MS20_LastSeed,1,MS20_Nof_BAT);
  MS20_LambdaN=dmatrix(MS20_FirstSeed,MS20_LastSeed,1,MS20_Nof_BAT);
  MS20_NofLD=5;
  MS20_LD=ul3tensor(MS20_FirstSeed,MS20_LastSeed,
		    1,MS20_Nof_BAT,1,MS20_NofLD
		   );



  /* output file */
  sprintf(fn,"rndma_lambda1_k1.dat");
  MS20_OFP=fopen(fn,"w");
  if(!MS20_OFP){
        fprintf(stderr,"ERROR: couldn't open the file\n  %s\nfor appending",fn);
	MY_EXIT;
  }

  fprintf(MS20_OFP,"\n");
  fprintf(MS20_OFP,"\n\n\n");
  fprintf(MS20_OFP,"# extremal eigvals (+ inv.part.ratios) and largest degrees of graphs\n");
  fprintf(MS20_OFP,"\n");
  fprintf(MS20_OFP,"# size\t\tseed\tlambda_1\tlambda_2\tlambda_N");
  for(ii=1;ii<=MS20_NofLD;ii++){
          fprintf(MS20_OFP,"\tk%ld",ii);    
  }
  fprintf(MS20_OFP,"\tipr_1\t\tipr_2\t\tipr_N");
  fprintf(MS20_OFP,"\n");
  fprintf(MS20_OFP,"\n");
  fflush(MS20_OFP);
}

/*-------------------------------------------*/

void MS20_Init_Inside( int bat_index ){

  switch(Model){
  case 0: default:{
          MS20_nVertex =   (unsigned long)10000*ER_nVertex_div10000
	                + (unsigned long)ER_nVertex_mod10000;
	  MS20_nEdge =     (unsigned long)10000*ER_nEdge_div10000 
	                + (unsigned long)ER_nEdge_mod10000;
	  break;
  }
  case 1:{
          BA_T = MS20_BAT[bat_index];
	  MS20_nVertex = (unsigned long)BA_M0 + BA_T;
	  MS20_nEdge = (unsigned long)BA_M * BA_T;
	  break;
  }
  case 2: case 20:{
          UC_NTryLimit =   (unsigned long)10000*UC_NTryLimit_div10000 
	                 + (unsigned long)UC_NTryLimit_mod10000;
          MS20_nVertex =   (unsigned long)10000*UC_nVertex_div10000 
	                + (unsigned long)UC_nVertex_mod10000;
	  break;
  }
  case 3:{
          MS20_nVertex = MS20_BAT[bat_index];
	  MS20_nEdge = MS20_nVertex * SW_k/2;
	  break;
  }
  }


  MS20_Degree=lvector(1,MS20_nVertex);
}

/*-------------------------------------------*/

void MS20_Run( int seed, int bat_index, 
	       double **lambda1, 
	       double **lambda2, 
	       double **lambdaN, 
	       unsigned long *degree,
	       unsigned long *ld,
	       double *ipr1, double *ipr2, double *iprN
	       ){

  unsigned long *iarr;
  int i, niter[1];
  double *lambda, *beta, *error, sumomegaabs[1], *ipr;
  lambda = dvector(1,Lanz_M);
  beta = dvector(1,Lanz_M);
  error = dvector(1,Lanz_M);
  ipr = dvector(1,Lanz_M);
  /*
   * bat is the "t" parameter of the BA model 
   * bat_index: indexes the MS20_BAT array
   * ipr: inverse participation ratio (of eigenvectors)
   */
  


  /* generating matrix */
  GenerateMatrix( seed, Model, 
		  MS20_nVertex, MS20_nEdge, 
		  BA_M0, BA_M, MS20_BAT[bat_index],
		  UC_Gamma, MS20_nVertex, &MS20_UC_nEdge,
		  UC_minDegree, UC_pN, UC_NTryLimit,
		  MS20_nVertex, SW_k, SW_p,
		  degree
                );


  /* largest eigvals, and the inverse participation ratios of their
   * eigenvectors 
   */
  /* compute */
  GetEigVals( MS20_nVertex, 0, 0.0, 0, 0.0,
	      degree, niter, lambda, beta, error, sumomegaabs, ipr
	      );
  /* save */
  lambda1[seed][bat_index] = lambda[1];
  lambda2[seed][bat_index] = lambda[2];
  *ipr1 = ipr[1];
  *ipr2 = ipr[2];



  /* smallest eigvals, and the inv.part.ratio of its eigenvector */
  /* compute */
  GetEigVals( MS20_nVertex, 4, 0.0, 0, 0.0,
	      degree, niter, lambda, beta, error, sumomegaabs, ipr
	      );
  /* save */
  lambdaN[seed][bat_index] = lambda[1];
  *iprN = ipr[1];


  /* save largest degrees */
  iarr=lvector(1,MS20_nVertex);
  indexx_ul(MS20_nVertex,degree,iarr);
  for(i=1;i<=MS20_NofLD;i++){ 
          ld[i] = degree[iarr[MS20_nVertex+1-i]];
  }



  free_lvector(iarr,1,MS20_nVertex);
  free_dvector(lambda,1,Lanz_M);
  free_dvector(beta,1,Lanz_M);
  free_dvector(error,1,Lanz_M);
  free_dvector(ipr,1,Lanz_M);
}

/*-------------------------------------------*/

void MS20_EvalSave_Inside( int seed, int bat_index, 
			   double **lambda1, 
			   double **lambda2, 
			   double **lambdaN, 
			   unsigned long ***ld,
			   double ipr1, double ipr2, double iprN
			   ){

  int ii;


  /* write largest eigval and largest row sums into file */
  fprintf(MS20_OFP,"%.7E\t%d\t%.7E\t%.7E\t%.7E",
	  (double)MS20_nVertex,seed,lambda1[seed][bat_index],
	  lambda2[seed][bat_index],lambdaN[seed][bat_index]);
  for(ii=1;ii<=MS20_NofLD;ii++){
          fprintf(MS20_OFP,"\t%ld",ld[seed][bat_index][ii]);
  }
  fprintf(MS20_OFP,"\t%.7E\t%.7E\t%.7E",ipr1,ipr2,iprN);
  fprintf(MS20_OFP,"\n");
  fflush(MS20_OFP);
}

/*-------------------------------------------------*/

void MS20_Clean_Inside(){

  unsigned long i,*nei_ncl,*nei_nch;
  nei_ncl=lvector(1,MS20_nVertex);
  nei_nch=lvector(1,MS20_nVertex);
  
  for(i=1;i<=MS20_nVertex;i++){
          nei_ncl[i] = 1;
          if(MS20_Degree[i]==0){ nei_nch[i] = 1; }
	  else{ nei_nch[i] = MS20_Degree[i]; }
  }
  free_lmatrix_vrl(NEI,1,MS20_nVertex,nei_ncl,nei_nch);
  free_lvector(MS20_Degree,1,MS20_nVertex);
  free_lvector(nei_ncl,1,MS20_nVertex);
  free_lvector(nei_nch,1,MS20_nVertex);
}

/*-------------------------------------------*/

void MS20_EvalSave_AtEnd( double **lambda1, 
			  double **lambda2, 
			  double **lambdaN, 
			  unsigned long ***ld 
			  ){

  fflush(MS20_OFP);  
  fclose(MS20_OFP);
}

/***************************************************/

#undef MY_EXIT
#undef MY_STRLEN
#undef MY_SQR
#undef MY_ABS
#undef MY_MAX
#undef MY_MIN

